#define VERSION "2.6a"

#if defined(LINKCELL) && defined(FREEBC)
#error "LINKCELL not supported for FREEBC"
#endif

#include "ground.h"
#define SDS
#include "alloc.h"
#include <time.h>

#include "options.h"
#include "optlist.c"

#include <signal.h>

#include "varfile.h"
#include "statics.h"
#include "linregr.h"

#include "rndgen.h"
#include "simglob.h"

#ifndef PARALLEL
#define PARALLEL 0
#endif /*# PARALLEL */

#include "asksig.h"

#if defined(Hamilton) == defined(Lagrange)
#error "Only one of {Hamilton Lagrange} may be #defined"
#endif /*# defined(Hamilton) == defined(Lagrange) */

#include "units.h"
#include "simgear.h"
#include "simcg.h"

#include "interpot.h"
#include "simdef.h"

/*** some additional global variables ***/

#ifdef SHAKE
int shake=0; /* Verlet integrator with possible SHAKE constraints */
#endif /*# SHAKE */

/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                       Ewald summation & erfc                          %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef COULOMB
#include "ewald.h"
#include "elst.h"
#endif /*# COULOMB */

#include "simpot.h"
#include "siminit.h"
#ifdef DIHEDRAL
/* WARNING this is something very obsolete... */
#include "dihedral.h"
#endif /*# DIHEDRAL */
#include "simmeas.h"

#include "setss.h" /* widomrdf() needed */

#include "norm.h"
#ifdef LINKCELL
#include "linklist.h"
#endif

#include "simils.h"
#include "rhs.h"
#include "constrd.h"
#include "cputime.h"

#include "sfdx.h"

#ifdef MARK
#include "mark.h"
#endif /*# MARK */

static FILE *lock;
static void removelock(void) /*********************************** removelock */
/*
  removes the lock file SIMNAME.loc
  registered by atexit() => to be called at exit()
*/
{
  if (lock) fclose(lock);
  remove(Fn("loc"));
}

static int justnow(double dt,double half) /************************* justnow */
{
  dt=fabs(dt); /* important! */
  if (dt==0) return 0;

  return (int)((t+half)/dt+1e4) > (int)((t-half)/dt+1e4);
}

static int SI=1; /* 1 if SI units, 0 if kcal/mol etc. */
static double SIE=1;

static void StaAddE(const char *name,double x) /******************** StaAddE */
{
  char nm[128];

  strcpy(nm,name);
  if (SI) strcat(nm," [J/mol]");
  else strcat(nm," [kcal/mol]");

  StaAdd(nm,x/SIE);
}

int main(int narg,char **arg)
{
  static int
    noint=10,     /* number of steps in 1 cycle = frequency of measurements */
    no=10000,     /* # of cycles in 1 main cycle */
    nomax=0,      /* max # of cycles from start of measurements (init>=1) */
    initno=0,     /* option -NUMBER (overrides no in the 1st cycle) */
    quit=0,       /* quit=1 causes program stop */
    sort=0,       /* sort by CM, coordinate: 1=x, etc. */
    loadreversed=0, /* 1 if to reverse endian on loading cfg */
    fracpass=0,   /* turned off by any option q and in further cycles */
    pass=0,
    CPnbit=0,     /* resolution (1/2^CPnbit) for packed convergence profiles */
    rhoreached,   /* switch to check whether rho was reached */
    i,nsteps,initvel=0,plbopened=0,issta=0,frame0,expectedbuffer;

  static volatile int
    icyc=0,iint=0;

  static double
    stop=0,       /* max time from simulation start (init>=2) */
    Emax=1e4,     /* max energy of inserted molecule - see initcfg */
    drmax=0.125,  /* max displacement in one integration step (init>2 only)
                     drmax<0: applies for any init */
    bulkmodulus=0, /* for NPT */
    maxscale=1.03,/* max scaling of box, applies with tau.P, tau.R, (tau.rho) */
    pins=0.01,    /* limit probability of insertions - see initcfg */
    initrho=0,    /* initial density - see initcfg; 0=rho */
    dV=0,         /* volume-change and surface-change, cf. constrd.dV */
    DT,           /* cycle in s */
    minmass=0,    /* to check h */
#ifdef POLAR
    epspq=0.8,    /* epsp enhancement factor for simulation (re)start
                    (close to 1:more, epspq<=0:off) */
    epsp0=0,      /* copy of epsp */
#endif /*# POLAR */
    LJcutoff=-3;  /* Lennard-Jones cutoff - see comments in simdef.c! */

  static struct {
    double
      grid,       /* intervals/1A; negative:s-s rdf's merged; 0:none */
      cutoff;     /* range of rdf */
    int
      onefour;    /* 1 if 1-4 are to be included to RDF's, 0 otherwise */
  } rdf = {0,0,1};

  static struct {
    double
      prt, /* frequency of printing t,T,E,... to out [in ps] (default=every) */
      cfg, /* save configuration frequency [in ps] - see savecfg */
      plb; /* frequency of saving playback file */
  } dt = { 1e-4,0,0 };

  static vector shift; /* added to cfg. before run, then :=0 */
  static vector Eelst; /* external elst field, in V/m */

#ifdef WIDOM
  static struct {
    int spreal; /* for scaled particle (id change): the real species */
    int sp; /* species to insert or be changed to virtually */
    int n; /* # of attempted insertions per cycle [WALL: and slab]; negative: excl. wall-mol pot. */
#ifdef SLAB
    int mode; /* also symmetrized profile */
    double z0,z1,dz; /* define slabs */
#endif /*# SLAB */
    double corr; /* cutoff correction without division by V */
  } widom={-1,0,0,2};
#endif /*# WIDOM */

#ifdef SLAB
  static struct {
    double grid; /* histogram bins/A */
    double zmax; /* range is [0,zmax] */
    int slab;    /* 1: calculate surface tension incl. cutoff corr.
                    2: with dV, forces const-volume scaling */
    } densprof = { 0,0,0 };
#endif /*# SLAB */

#ifdef XSECTION
  static struct {
    double grid;   /* grid of the mesh, in points per A */
    double RvdW;   /* van der Waals radius of the test particle, additive */
    double Rscale; /* multiply van der Waals radii; default=2^(-1/6) */
    int mode;      /* 0..4, see sfdx.c */
    int freq;      /* how often to calculate, in # of cycles (1=every cycle) */
    int cp;        /* -1: averaged xsec in CP, >=0 : 1 mol value */
    int ns;        /* max # of sites incl. in Xsec calculations; -1=all */
    int sizelimit; /* max size of the grid array */
    int maxsize;   /* that one actually encountered */
  } xs={10.0,1.4,0.89089871814033927,0,0,0,-1,64*1024*1024,0};
#endif /*# XSECTION */

#ifdef RGYR
  static struct {
    int end[2];   /* indices of `ends' */
    int cp;       /* -1: averaged Rgyr and end-to-end in CP, >=0 : 1 mol.val. */
  } rg={{0,-1},0};
#endif   /*# RGYR */

#ifdef DIHHIST
  static struct {
    double res;  /* resolution, in deg */
    int grid;   /* # of intervals / 360 deg */
    int cp,dcp;
  } dih={0,0,-1,0};
#endif /*# DIHHIST */

#ifdef DISTURB
  double disturb[2];  /* random perturbation of cfg. - see disturb() */
#endif /*# DISTURB */

  vector Pscale={1,0,0};
  double RvdW=0; /* particle size to adjust: see tau.R, tau.i */

  double lastEtot,halftcyc,V;

  static char *sysname=NULL;
  char *mark;
  time_t starttime,stoptime,stop_start;
  static double rho=0; /* density [kg/m3] */
  static double refrho;
  static vector L; /* box size (in input data) */
  static double cutoff; /* cutoff (in input data) */
  static int corr=1; /* 1=cutoff corrections */

  /* support of strings like init="cont" in getdata */
  static struct keys_s initkey[]={
    {"continue",0}, {"cont",0},
    {"append",1},
    {"start",2},
    {"random",3},
    /*    {"rawrandom",4},*/
    {"slab",4},
    {"crystal",5}, {"lattice",5},
    {"asc",10},
    {"convert",999},
    {"plb",-1},
    {NULL,0}};
  static struct keys_s yesno[]={
    {"no",0}, {"false",0},
    {"yes",1},{"true",1},
    {"forever",-1}, /* for MC */
    {"debug",-1}, /* for quit */
    {"paste",-2}, /* for quit */
    {"shell",-4}, /* for quit */
    {NULL,0}};
  static struct keys_s pinskey[]={
    {"auto",0},
    {"sc",1},
    {"bcc",2},
    {"fcc",3},
    {NULL,0}};
  static struct keys_s thermostatkey[]={
    {"none",0},
    {"friction",T_FRICTION},{"Berendsen",T_BERENDSEN}, /* T_FRICTION=T_BERENDSEN */
    {"Nose",T_NOSE},{"Nose-Hoover",T_NOSE},
    {"Andersen",T_ANDERSEN},
    {"Maxwell",T_MAXWELL},
    {"AndersenCM",T_ANDERSEN_CM},
    {"MaxwellCM",T_MAXWELL_CM},
    {"tr",T_TR},
    {"in",T_IN},
    {"frictions",T_FRICTIONS},
    {NULL,0}};
  static struct keys_s sortkey[]={
    {"none",0},
    {"x",1}, {"-x",-1},
    {"y",2}, {"-y",-2},
    {"z",3}, {"-z",-3},
    {NULL,0}};
static struct keys_s rescalekey[]={
    {"CM",RESCALE_XYZ|RESCALE_CM},
    {"x",RESCALE_X},           {"xCM",RESCALE_X|RESCALE_CM},
    {"y",RESCALE_Y},           {"yCM",RESCALE_Y|RESCALE_CM},
    {"xy",RESCALE_X|RESCALE_Y},{"xyCM",RESCALE_X|RESCALE_Y|RESCALE_CM},
    {"z",RESCALE_Z},           {"zCM",RESCALE_Z|RESCALE_CM},
    {"xz",RESCALE_X|RESCALE_Z},{"xzCM",RESCALE_X|RESCALE_Z|RESCALE_CM},
    {"yz",RESCALE_Y|RESCALE_Z},{"yzCM",RESCALE_Y|RESCALE_Z|RESCALE_CM},
    {"xyz",RESCALE_XYZ},       {"xyzCM",RESCALE_XYZ|RESCALE_CM},
    {"xx",RESCALE_X|RESCALE_Y|RESCALE_XisY},{"xxCM",RESCALE_X|RESCALE_Y|RESCALE_XisY},
    {"xxz",RESCALE_XYZ|RESCALE_XisY},{"xxzCM",RESCALE_XYZ|RESCALE_XisY},
    {"X",RESCALE_X|RESCALE_PT},{"XCM",RESCALE_X|RESCALE_PT|RESCALE_CM},
    {"Y",RESCALE_Y|RESCALE_PT},{"YCM",RESCALE_Y|RESCALE_PT|RESCALE_CM},
    {"XY",RESCALE_X|RESCALE_Y|RESCALE_PT},{"XYCM",RESCALE_X|RESCALE_Y|RESCALE_PT|RESCALE_CM},
    {"Z",RESCALE_Z|RESCALE_PT},{"ZCM",RESCALE_Z|RESCALE_PT|RESCALE_CM},
    {"XZ",RESCALE_X|RESCALE_Z|RESCALE_PT},{"XZCM",RESCALE_X|RESCALE_Z|RESCALE_PT|RESCALE_CM},
    {"YZ",RESCALE_Y|RESCALE_Z|RESCALE_PT},{"YZCM",RESCALE_Y|RESCALE_Z|RESCALE_PT|RESCALE_CM},
    {"XYZ",RESCALE_XYZ|RESCALE_PT},       {"XYZCM",RESCALE_XYZ|RESCALE_PT|RESCALE_CM},
    {"XX",RESCALE_X|RESCALE_Y|RESCALE_PT|RESCALE_XisY},{"XXCM",RESCALE_X|RESCALE_Y|RESCALE_PT|RESCALE_XisY|RESCALE_CM},
    {"XXZ",RESCALE_XYZ|RESCALE_PT|RESCALE_XisY},{"XXZCM",RESCALE_XYZ|RESCALE_PT|RESCALE_XisY|RESCALE_CM},
    {NULL,0}};

  if (narg<2) {
    fprintf(stderr,"*** " PROJECT " *** V" VERSION " *** (c) J.Kolafa 1991-2010 ***\n");
    fprintf(stderr,"Call by:\n\
   %s [OPTIONS] SYSNAME SIMNAME [OPTIONS]\n",arg[0]);
    fputs("\
Input files:\n\
   SYSNAME.ble  force field, generated by blend\n\
   SIMNAME.def  defaults; if no SIMNAME.def, SYSNAME.def is tried\n\
   SIMNAME.get  input data (if -s then stdin)\n\
   SIMNAME.cpi  variables recorded in convergence profiles\n\
OPTIONS:\n"
#ifdef POLAR
"  -a% scale polarizabilities\n"
#endif
"  -b# interactive (-s): set beep level  batch: SIMNAME.stp=^C (interrupt+save)\n\
  -c# 1=corr. constraints/Gear by conjgrad, report: 2=before 4=mid 8=after [9]\n\
  -d# check pair distances: 1=once per job 2=per cycle 3=per step [0=never]\n\
  -e# record max # items in the SIMNAME.cp\n"
#ifndef FREEBC
  "  -f# diffusivity and structure factor from stored cfgs [init,init+#,..,no]\n"
#endif /*# FREEBC */
"  -g# input=SIMNAME.G#; -f -g# output=SIMNAME.S#"
#ifdef CLUSTERS
", no cluster analysis"
#endif /*# CLUSTERS */
"\n  -h4 refine tetrahedral angles\n\
  -i# ^C (kill -SIGINT) is: -1=kill 0=ask 1=next data 2=finish\n\
  -j# don't rotate randomly molecules while init, #>0: fix 1st # molecules\n\
  -k# keep atoms (listed in SIMNAME.fix) by force # K/A^2\n\
  -lGFAVC ASCII dump (Grad,Forces,Accel,Veloc,Cfg); 1:full prec., >1: dec.dig.\n\
  -m# {3-6} Gear + Lagrangian constrained dynamics, default=-m2=Verlet+SHAKE\n\
  -n  write SIMNAME.vlb (velocity playback)\n\
  -o- ignore loc-file (the same as -o0)\n\
  -pKPC predictor lengths and orders:\n\
    C=Verlet: velocity predictor length, Gear: constraint predictor order (-m#)\n\
    P=Polar method, 0:Car-Parrinello, 1:no prediction, 2:ASPC, 3:higher order\n\
    K=ASPC or similar polar predictor length\n\
  -q% init charge scaling (in %), #<0: make cfg neutral\n\
  -r# record cfgs to #-th derivative (by dt.cfg)\n\
  -s# interactive mode (#kB scroll buffer)\n\
  -t  detailed time measurements\n\
  -u# energy limit for vibrating bonds in K/A^2 [0=all bonds constrained]\n\
  -v# verbosity: 0=min or sum of 1,2,4,..\n\
  -w# write cfg: 0=never 1=after data set (;) 2=wait if not enough disk space\n\
  -x  disable optimized water models\n\
  -y# playback # molecules only, -y-# don't call molcfg, (df.=whole cfg)\n\
  -z# random number seed, 0=default=use time\n\
  -[### load cfg+make periodic copies   -[-1000 | -[-111 load with wrong N | L\n\
  -]# equalize masses in molecules (0=no, 100=all equal)\n"
#ifdef POLAR
"  -^#%% POLAR: #=shell=Drude charge (signed in e), %%=omegap (2 digits, in %)\n"
#endif /*# POLAR */
#ifdef LINKCELL
"  -@# LINKCELL: record cell occupancies (to # sites/cell)\n"
#endif /*# LINKCELL */
#if PARALLEL
"  -\\# PARALLEL: -\\# # of threads; 0 use NSLOTS (df.) <0 linked-cell: No.cell[0]\n"
#endif /*# LINKCELL */
         ,stderr);
    exit(0); }

  out=stderr;

  /*** option analysis ***/
  loop (i,1,narg)
    if (arg[i][0]=='-')
      if (arg[i][1]>='0' && arg[i][1]<='9') initno=atoi(arg[i]+1);
      else {
        getoption(arg[i])
        if (arg[i][1]=='q') fracpass++; }
#if CHECKHEAP==-2
    /* to debug memory problems: option +AllocRange */
    else if (arg[i][0]=='+') {
/*.....    AllocSizeLim=atoi(arg[i]+1);*/
      AllocTrace++;
      AllocRange=atoi(arg[i]+1); }
#endif /*# CHECKHEAP==-2 */
    else {
      if (!sysname) sysname=arg[i];
      else if (simname) ERROR(("third argument (not prepended by -), max two (SYSNAME and SIMNAME) allowed"))
      else simname=arg[i]; }

  if (option('p')<0) ERROR(("negative option -p no longer supported"))
  No.pred=option('p')%10;
  if (No.pred==9) No.pred=option('m');

  if (option('s')==1)
#ifdef DOS
    option('s')=31;
#else /*# DOS */
    option('s')=256;
#endif /*# DOS */

#if PARALLEL
  if (option('i')==-9) option('i')=1+(!option('s'));
  if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
#if PARALLEL==1 || PARALLEL==3
  if (sizeof(struct pll_global_s)%CACHELINE) DISASTER(("sizeof(pll_global_s)=%d is not an integer multiple of the cache line=%d\n\
*** adjust padding in `struct pll_global_s' in sim/interpot.h and recompile!",sizeof(struct pll_global_s),CACHELINE))
#endif
#else /*# PARALLEL */
  if (option('i')==-9) option('i')=2*!option('s');
#endif /*# PARALLEL */

#ifdef POLAR
  if (option('f')) {
    option('a')=0;
    prt("option -a0 set because of -f"); }
#endif /*# POLAR */

#ifdef Lagrange
#ifdef SHAKE
  shake=option('m')==2;
#else /*# SHAKE */
  if (option('m')<=2) Error("SHAKE not #defined and -m2");
#endif /*# SHAKE */
#endif /*# Lagrange */

  if (!shake && option('c')&1) epsc=0.05;

  /* install SIGINT (2) and SIGTERM (15) signal handler */
  if (option('i')>=0) {
    signal(SIGINT,asksig);
    /* SIGTERM not for MPI ...? */
    signal(SIGTERM,asksig); }

  simils.rotatefrom=abs(option('j'));
  if (option('j')<0) option('j')=0;

  initscroll(option('s')*1024);

  if (!simname && sysname) {
    prt("WARNING: SIMNAME not specified, SIMNAME=SYSNAME will be used.");
    simname=sysname; }

  if (simname) {
    lock=fopen(Fn("loc"),"r+");
    out=stdout;
    if (lock) {
      if (option('o')) ERROR(("Lock file \"%s\" exists\n\
(\"cook %s\" may be running or has not been stopped properly)",
            lastFn,simname))
      else
        fprintf(stderr,"WARNING: Lock file exists but is ignored because of -o0\n"); }
    else
      lock=fopen(Fn("loc"),"w");

    atexit(removelock);
    remove(Fn("sig"));

    if (option('s')) {
      if (option('b')<0) option('b')=1; /* default beep level */
      out=stdout; }
    else {
      char *ext=option('f') ? "sfd" : "prt";
      char eee[12]="S";

      if (option('g') && option('f')) {
        ext=eee; sprintf(ext+1,"%d",option('g')); }
      else
        backup(ext);
      out=fopen(Fn(ext),"wt"); } }

  if (out==NULL) Error("open .prt");

  prts("*** " PROJECT " *** V" VERSION " *** (c) J.Kolafa 1991-2010 ***");

#ifdef Lagrange
  prts_("*** Lagrange");
#else /*# Lagrange */
  prts_("*** Hamilton");
#endif /*# Lagrange */

#ifdef FREEBC
  prts_(" *** FREEBC");
#endif /*# FREEBC */

#ifdef POLAR
  prts_(" *** POLAR");
#endif /*# POLAR */

#if PARALLEL
  prt_(" *** PARALLEL=%d",PARALLEL);
#endif /*# PARALLEL */

#ifdef REDUNITS
  prts_(" *** REDUNITS");
#endif /*# REDUNITS */

  { /* options, sex, environment */
    int Int=1,i;
    char *end=(char*)&Int;
    char *host=getenv("HOSTNAME");
    char *user=getenv("USERNAME");

    if (!host) host=getenv("HOST");
    if (!host) host="UNKNOWN";
    if (!user) user=getenv("USER");
    if (!user) user="UNKNOWN";

    prt(" *** %s endian s%d i%d l%d p%d ***",
        *end==1?"little":*end==0?"BIG":"?",
        sizeof(short),sizeof(int),sizeof(long),sizeof(void*));

    loop (i,0,narg) {
      prt_("%s%c",arg[i]," \n"[i==narg-1]);
      fprintf(stderr,"%s%c",arg[i]," \n"[i==narg-1]); }
    prt("HOST=%s USER=%s PWD=%s",host,user,getenv("PWD"));
    fprintf(stderr,"HOST=%s USER=%s PWD=%s\n",host,user,getenv("PWD"));
    prt("Edited arg list:");

    if (sysname) prt_("%s ",sysname);
    if (simname) prt_("%s",simname);
    if (simname || sysname) _n

    loop (i,0,32) {
      if (i==17) _n
      prtoption(i,optionlist) }
    _n
  }

  if (!simname || !sysname) ERROR(("no SIMNAME specified"))

  /* remove the stp-file so that the simulation is not immediately stopped */
  remove(Fn("stp"));
  _n

#if PARALLEL
  if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
#endif /*# PARALLEL */

  rndinit(7L/*low quality enough*/,option('z'));

#ifdef COOK
  /* see also mainble.c */
  mark=simname; simname=sysname; /* swap sysname/simname temporarily */
  in=fopen(Fn("ble"),"rt");
  simname=mark; /* normal sysname/simname again */
  if (!in) Error("open system ble-file");
  minmass=readblend();
  fclose(in);
#else /*# COOK */
  convertsiteunits(); /* just writes info #ifdef REDUNITS */
#endif /*# COOK */

  /* open sysname.def - to be read in pass 0 below */
  if ( !(in=fopen(Fn("def"),"rt")) ) {
    /* simname.def not found: trying sysname.def */
    char *swapname=simname;

    simname=sysname; /* swap sysname/simname temporarily */
    in=fopen(Fn("def"),"rt");
    simname=swapname; /* normal sysname/simname again */
    if (!in) ERROR(("no %s.def nor %s.def found",sysname,simname)) }

  while (sig<2) { /* <<< run <<< */
    sig= -1;

    checkranges(1);

    init=0; /* default in every pass */

    /* getdata : twice for the 1st time to read .def */
    do {

#ifdef COOK
      int *N,i;
      if (pass==0) {
        alloc(N,nspec*sizeof(int));
        loop (i,0,nspec) N[i]=spec[i]->N; }
#endif /*# COOK */

      if (pass==1) {
#ifdef COOK
        loop (i,0,nspec) spec[i]->N=N[i];
        free(N);
        if (option('k')>0) initfix(Fn("fix"));
        initNo(); /* now we know everything to call it */

        checkfixed();
#endif /*# COOK */
        /* close .def file, open .get/.G# file or use stdin */
        fclose(in);
        if (option('s')) in=stdin;
        else {
          char ext[12]="get";

          if (option('g')) sprintf(ext,"G%d",option('g'));
          in=fopen(Fn(ext),"rt"); }
        if (in==NULL) ERROR(("open %s",lastFn)) }

      if (pass>=1) {
        _n putline('%',79);
        if (option('s')) prt("Enter data in form `VARIABLE=EXPRESSION\', `?\' for help, `;\' to start simulation:"); }

      getdata
        /* auxiliary variables for calculations */
        static double
          aux=0, a=0, b=0, c=0, x=0, y=0, z=0;
        static int
          i=0,j=0,k=0,n=0;

#ifdef FREEBC
        get(center.K) get(center.r0)
#else
        get(el.epsk) get(el.epsr)
        get(el.alpha) get(el.kappa)
        get(el.epsinf) get(el.test) get(el.diag) get(el.diff)

        getvec(center.K,,DIM)
        getvec(center.r0,,DIM)
#ifdef SLAB
        getvec(center.n,,NCENTER)
        getvec(center.ns,,NCENTER)
        getvec(center.z,,NCENTER)
        getvec(center.Kz,,NCENTER)
        getvec(center.z0,,NCENTER)
        getvec(center.z1,,NCENTER)
#endif
#endif /*# FREEBC */
        getvec(center.cmK,,DIM)
        get(center.cmn)
        getvec(Eelst,,DIM)
#include "simget.c"
        get(dV)
        getkey(SI,yesno)
        getkey(init,initkey)
        get(mirror) getkey(sort,sortkey)
        getkey(pins,pinskey) get(initrho) getkey(MC,yesno)
        get(E) get(rho) getvec(L,,DIM)
        getvec(shift,,DIM)
        get(tau.E) get(tau.rho)
        getkey(rescale,rescalekey)
        get(tau.T) get(T) get(initvel)
        getkey(thermostat,thermostatkey)
        get(T_tr_in)
        get(tau.P) get(P) get(maxscale) get(bulkmodulus)
        get(tau.R) get(tau.i)
#ifdef DISTURB
        getvec(disturb,,2)
#endif /*# DISTURB */
        get(omega) get(omegac)
        get(eps) get(epsc)
        get(no) get(noint) get(h)
        get(nomax) get(stop)
        get(dt.prt) get(dt.cfg) get(dt.plb)
        get(lag.err) get(lag.n)
        get(lag.cond)
#if PRESSURETENSOR&PT_OFF
        get(lag.visc)
#endif
        get(Emax) get(drmax)
        get(rdf.grid) get(rdf.cutoff) get(rdf.onefour)
        get(el.grid) get(el.minqq)
        get(cutoff) get(LJcutoff) get(corr)
        getkey(loadreversed,yesno)
        get(CPnbit)
#ifdef POLAR
        get(epsp) get(epspq) get(omegap) get(maxitp) get(epspx)
        get(tau.dip)
#endif /*# POLAR */
#ifdef SHEAR
        get(shear)
#endif /*# SHEAR */
#ifdef LINKCELL
        getvec(No.cell,,DIM)
        get(No.percell)
        get(over14) get(box.rmin)
#endif /*# LINKCELL */
#ifdef WALL
        /* structure wall defined in simdef */
        get(wall.rho) get(wall.minz) get(wall.g) get(wall.scalez)
        get(wall.n) get(wall.rep)
#endif /*# WALL */
        get(No.first)
        get(drift) get(conserved)
        getkey(quit,yesno)
#ifdef WATERPLUS
        {
#include "watercut.h"
          get(cut.from) get(cut.to) }
#endif /*# WATERPLUS */
        if (quit==-4) {
          /* the shell; GCC/EMX starts always cmd.exe */
          int i;

          fprintf(stderr,"\n *** Type `exit' to return to %s! ***\n",arg[0]);
#ifdef DOS
          i=system("command.com");
#elif defined(__EMX__)
          i=system("cmd.exe");
#else /* UNIX (and other?) */ /*# DOS */
          i=system(getenv("SHELL")?getenv("SHELL"):"bash");
#endif /*# DOS */
          prt("shell return code=%d",i);
          quit=0; }

        get(cache)
#ifdef DIHHIST
        get(dih.res) get(dih.grid) get(dih.cp) get(dih.dcp)
#endif /*# DIHHIST */

#ifdef WIDOM
        get(widom.sp) get(widom.spreal) get(widom.n)
#ifdef SLAB
        get(widom.z0) get(widom.z1) get(widom.dz) get(widom.mode)
#endif /*# SLAB */
#endif /*# WIDOM */

#ifdef SLAB
        get(densprof.grid) get(densprof.zmax) get(densprof.slab)
#endif /*# SLAB */

#ifdef XSECTION
        get(xs.grid) get(xs.RvdW) get(xs.Rscale) get(xs.mode) get(xs.freq)
        get(xs.cp) get(xs.ns) get(xs.sizelimit)
#endif /*# XSECTION */

#ifdef RGYR
        getvec(rg.end,,2) get(rg.cp)
        get(lag.v)
#endif /*# RGYR */

        /* auxiliary variables */
        get(aux)
        get(a) get(b) get(c)
        get(x) get(y) get(z)
        get(i) get(j) get(k)
        get(n)

        checkdata
        enddata
      _n

      onefourinrdf=rdf.onefour;

#ifdef CLUSTERS
      /* reads both .def and .get */
      if (option('f') && !option('g')) readclusterdef(in);
#endif /*# CLUSTERS */

      if (quit>0) goto TheEnd;
      } while (pass++==0);

#if 0
    if (option('y')<0) {
      static int pass;

      if (!pass++)
      WARNING(("The meaning of option -y%d changed. Since V2.4g:\n\
***   -y-# = record # molecules only (without calling molcfg)\n\
***   -yBIG = record all molecules as usual (now the default)",option('y'))) }
#endif

    if (option('y')>0 && (init<0 || init>1) && !option('f')) makemolgol(option('y'));
    init_append=0;

    if (tau.E!=0 && thermostat) {
      WARNING(("tau.E!=0 and thermostat: thermostat turned off"))
      thermostat=0; }

    setdrift();

    prt("minimum atom mass=%g g/mol => suggested timestep h in (%.4f,%.4f) ps",
        minmass, 0.0005*sqrt(minmass), 0.0015*sqrt(minmass));
    if (h<0.0005*sqrt(minmass))
      prt("WARNING: the timestep h=%g is probably unnecessarily short",h);
    if (h>0.0025*sqrt(minmass))
      WARNING(("the timestep h=%g is probably too long\n\
*** bad energy conservation or crash may occur",h))
    if (h>0.0015*sqrt(minmass)) prt("\
WARNING: the timestep h=%g is probably too long\n\
         bad energy conservation or crash may occur",h);

    if (rho==0 && L[0]*L[1]*L[2]==0)
      if (init>=3) ERROR(("undefined box: set box (L[0],L[1],L[2]) or density (rho)"))

    constrd.ncoord="\0\1\1\2\1\2\2\3"[rescale&RESCALE_XYZ];

    underline("rescaling of the simulation box");
    prt("rescale = %d = 0x%x = %s (active only if tau.rho or tau.P)",rescale,rescale,int2sumbin(rescale));
#if (PRESSURETENSOR&PT_ANY) != PT_ANY
    if (rescale&RESCALE_PT)
      ERROR(("Pressure tensor components requested but this version of cook not\n\
*** compiled so. Use at least PRESSURETENSOR=3"))
#endif
    if (rescale&RESCALE_XYZ) {
      prt("configuration rescaling in these %d coordinate(s): %s\n\
based on %s",
        constrd.ncoord,
        prtxyz(rescale&RESCALE_XYZ),
        rescale&RESCALE_PT?"pressure tensor components":"isotropic virial pressure"); }
    else {
      prt("no configuration rescaling allowed, dV=tau.P=tau.rho=0 set\n\
*** (check rescale if this is not what you want)");
      dV=tau.P=tau.rho=0; }

    if (dV<0) {
      ERROR(("dV<0 (molecule-based rescaling) no longer supported, use dV>0 and rescale=8|(more options)"))
      dV=fabs(dV);
      rescale|=RESCALE_CM; }

    prt("configuration rescaling will be based on %s",
      rescale&RESCALE_CM ? "molecular centers-of-mass" : "sites");
    _n

    constrd.dV=dV;
    if (dV) {
      if (tau.P<0)
        ERROR(("every-cycle barostat (tau.P<0) is not supported with pressure by virtual volume change (dV)\n\
*** use tau.P>0"))
      else if (tau.P>0)
        WARNING(("the friction barostat will use the virtual volume change-based pressure")) }
    constrd.PdVname = rescale&RESCALE_CM ? "PdVmol [Pa]" : "PdVatom [Pa]";

    if (thermostat>=T_ANDERSEN && thermostat<=T_MAXWELL_CM) {
#if VERLET==1 || VERLET==2
      WARNING(("kinetic energy inacccurate with Andersen/Maxwell thermostat\n\
*** recompile with VERLET=3 (recommended) or 0 (less accurate)"))
#endif
      if (option('m')>2)
        WARNING(("Gear integration inacccurate with Andersen/Maxwell thermostat")) }

    if (tau.T && !thermostat)
      prt("WARNING no thermostat: nonzero tau.T=%g ignored\n\
(if you mean friction thermostat as in old cook, use thermostat=\"friction\")",tau.T);
    if (tau.T==0 && thermostat)
      WARNING(("thermostat with tau.T=0 will cause division by zero"))

    /* total charge check delayed to be interactive */
    if (fracpass==0) {
      int sp, nfrac=0;
      double charge;

      fracpass++;
      loop (sp,0,nspec) {
        charge=spec[sp]->charge/electron;
        if (fabs(fmod(fabs(charge+0.5),1)-0.5) > 1e-5) {
          nfrac++;
          prt("charge =%10.6f e of species %d is fractional",charge,sp); } }
      if (nfrac)
        WARNING(("\
%d molecule(s) with fractional charges\n\
(you may need blend -_PREC, check also mol-file)",nfrac)) }

    if (initno) {
      prt("no overriden by option -%d",initno);
      no=initno;
      initno=0; }

    underline("units");
#ifdef REDUNITS
    prt("! WARNING: reduced units used even if real units printed !");
#else /*# REDUNITS */
    if (SI) {
      SIE=1;
      prt("energy in statistics and convergence profile in J/mol"); }
    else {
      SIE=kcal;
      prt("energy in statistics and convergence profile in kcal/mol"); }
    prt("if no units are given (as for Etot), energy is in Kelvin");
#endif /*# REDUNITS */

    sig=0;

#ifdef DIHHIST
    if (dih.cp<-1) dih.cp=-1;
#endif /*# DIHHIST */

#ifdef COULOMB
#if COULOMB<=-1
    if (el.kappa<=0) if (!option('f')) WARNING(("el.kappa=%g is suspicious for Ewald",el.kappa))
#else /*# COULOMB==-1 */
    if (el.kappa) WARNING(("el.kappa=%g is suspicious for cutoff electrostatics",el.kappa))
#endif /*# COULOMB==-1 */
#endif /*# COULOMB */

    GearOrder=option('m');

#ifdef POLAR
    if ((option('p')/10)%10==0) if (tau.dip==0)
      ERROR(("Car-Parrinello-like method and tau.dip=0"))
    if (abs(option('^'))>99) option('^')=abs(option('^'))%100;
    if (omegap==-9) { /* omegap not specified in data */
      if (option('^'))
        omegap=option('^')/100.;
      else
        /* no -^ nor omegap: defaults
           - for -m2 (ASPC) the default is postponed to pred2.c
           - for -m>2 (full-iteration) the default is omegap=0.95
        */
        if (option('m')!=2) omegap=0.95; }
    else {
      /* omegap specified in data */
      option('^')=100*omegap+0.5; /* not to be set to default in pred2.c */ }
#endif /*# POLAR */

    putkey(init,initkey)
    if (init>=3 && init<10) {
      putkey_(pins,pinskey) put2(initrho,Emax) }
    put3(rho,tau.rho,el.epsinf)
    put2(E,tau.E)
    put_(T) put_(tau.T) putkey(thermostat,thermostatkey)
    if (thermostat>=10) put(T_tr_in)
    put2(P,tau.P)
    if ((tau.P!=0 || tau.rho!=0) && tau.R!=0)
      ERROR(("intermolecular parameter adjustement requires NVT"))

#ifdef WALL
    put3(wall.n,wall.rho,wall.minz)
    put2(wall.g,wall.scalez)
#endif /*# WALL */
    if (rdf.grid) put2(rdf.grid,rdf.cutoff)

#ifdef DIHHIST
    if (dih.res!=0) {
      if (dih.grid)
        WARNING(("max one of dih.grid, dih.res may be specified (dih.grid will be ignored)"))
      dih.grid=(int)(360./dih.res+0.5);
      if (fabs(360./dih.res/dih.grid-1)>1e-6)
        WARNING(("dih.res rounded to %f",360./dih.grid));
      dih.res=0; }
#endif /*# DIHHIST */

/* rdf is parallelized with PARALLEL==1, dih is serial (intramolecular) */
#if PARALLEL==3
    if (rdf.grid) {
    prt("WARNING: rdf not parallelized, all measurements serial (check noint)");
    No.measureserial++; }
#ifdef DIHHIST
    if (dih.grid) {
      WARNING(("WARNING: dih not parallelized, all measurements serial (check noint)"))
      No.measureserial++; }
#endif /*# DIHHIST */
#endif /*# PARALLEL==3 */

    if (tau.P*tau.rho) {
      WARNING(("both tau.rho and tau.P: tau.P replaced by zero"))
      tau.P=0; }

    /* put(omega) */
    if (!(option('c')&1)) put_(omegac)
    put2(epsc,eps)
    put3(no,noint,h)
    DT=noint*h*timeunit;
    prt("cycle (of noint MD steps): dt=%g s",DT);
    put3(dt.prt,dt.cfg,dt.plb)
    put2(lag.err,lag.n)
#if PRESSURETENSOR&PT_OFF
    put_(lag.visc)
#endif
#ifdef RGYR
    put_(lag.v)
#endif
    put(lag.cond)
    StaSet(DT,lag.err,2,lag.n);
    put3(cutoff,rdf.cutoff,LJcutoff)
    if (LJcutoff==0) WARNING(("LJcutoff=0 (no Lennard-Jones interaction)"))

#ifdef XSECTION
    prt("\
xs.grid=%g  xs.RvdW=%.8f  xs.Rscale=%.8f\n\
xs.mode=%d  xs.freq=%d xs.ns=%d  xs.sizelimit=%d",
xs.grid,    xs.RvdW,      xs.Rscale,
xs.mode,    xs.freq,   xs.ns,    xs.sizelimit);
    if (xs.sizelimit>AllocSizeLim) {
      WARNING(("xs.sizelimit=%d: are you sure?",xs.sizelimit))
      AllocSizeLim=xs.sizelimit; }
#endif /*# XSECTION */
    put(cache)

    /*** repeatable initializations ***/
    ralloc(mark,1);

#ifndef COOK
    initspecies(); /* init spec + calculates No */
#endif /*# COOK */

    underline("box size setting and determination of cutoff");
    refrho=setL(L,rho);

    prt("reference box = %.8f %.8f %.8f A",VARG(L));
    prt("reference rho = %.4f kg/m3",refrho);
    prt("> The reference box and density are derived from input rho and L[].\n\
> They are used for setting cutoff and Ewald parameters el.alpha, el.kappa\n\
> (if these are not given explicitly).\n\
> The actual box may differ:\n\
> * in NPT simulation (see tau.P and rescale)\n\
> * if other box is loaded and change is not allowed (see tau.rho and rescale)");

    VV(box.L,=L)
    if (init>=0 && init<3 || option('f'))
      loadcfg(-1,loadreversed,option('['),&RvdW); /* 1st pass */
    else if (init<0)
      readplayback(-init,loadreversed,0); /* L only */
    else if (init>10)
      readasc(0,0); /* L only */
    else {
      if (initrho && fabs(initrho/refrho-1)>5e-16) {
        prt("Initial density (initrho) differs from the reference density.");
        setL(box.L,initrho); }
      else
        prt("Initial density is given by rho/L specified"); }

    /* now: box.L=actual box loaded or to be initialized,
            L=reference or target box */

    replaceL(L); /* special for option -[-XYZ */
    rhoreached=0;

    if (fabs(box.L[0]/L[0]-1)>5e-16
     || fabs(box.L[1]/L[1]-1)>5e-16
     || fabs(box.L[2]/L[2]-1)>5e-16) {
      prt("WARNING: The actual (loaded/initial) box size differs from the reference size.");
      if (tau.P) {
        if (box.L[0]/L[0]-1<0 || box.L[1]/L[1]-1<0 || box.L[2]/L[2]-1<0)
          prt("WARNING: tau.P specified and the reference box is larger than actual\n\
(this is often bening, but sometimes may compromise accuracy of electrostatics)");
        }
      else if (tau.rho==0 || (rescale&7)==0) {
        prt("WARNING: Box size change is not allowed because tau.rho==0 or (rescale&7)==0\n\
==> the reference box will be replaced by the actual box (and kept constant)");
        VV(L,=box.L) }
      else if (tau.rho) {
        rhoreached++;
        prt("Since tau.rho is specified, the box will be gradually rescaled until the\n\
reference box is reached."); } }
    if (tau.P)
      prt("Since tau.P is specified, the box size will fluctuate; the reference box\n\
will be used for parameter setting.");

    VV(box.Lh,=0.5*box.L)

    box.cutoff=initcutoff(cutoff,L); /* set cutoff; needs No, L */

    if (rdf.cutoff>box.cutoff)
      WARNING(("rdf.cutoff=%g > cutoff=%g",rdf.cutoff,box.cutoff))
    if (rdf.cutoff==0) rdf.cutoff=fmin(box.cutoff,10);
    initrdf(rdf.grid,rdf.cutoff,Fn("s-s"));
#ifdef DIHHIST
    /* only histograms initialized - control structures already initialized */
    initdihhist(dih.grid);
#endif /*# DIHHIST */
#ifdef HARMONICS
    initharmonics(rdf.grid,rdf.cutoff,T,refrho);
#endif /*# HARMONICS */

    constrd.mode=rescale+RESCALE_L; /* force box rescale */
#ifdef SLAB
    if (densprof.slab && !densprof.grid)
      WARNING(("densprof.slab specified, but densprof.grid not given:\n\
*** surface tension and cutoff corrections will not be calculated\n"))
    if (densprof.slab&2) constrd.mode|=RESCALE_SLAB;
    initdpr(densprof.grid,densprof.zmax,Fn("dpr"));
#endif /*# SLAB */
    initss(LJcutoff);
    initpot();
    initmolecules(); /* here cfg allocated */
    initconstrit(omegac);
#ifdef WALL
    initwall();
#endif /*# WALL */

    /* external forces */
#ifdef FREEBC
    put2(center.K,center.r0)
#else /*# FREEBC */
    {
      int k;

      center.on=0;

      loop (k,0,DIM) {
        if (center.K[k]!=0) center.on |= 2;
        center.K2[k]=2*center.K[k];
        if (center.r0[k]!=0 && center.K[k]==0)
          WARNING(("center.r0[%d]=%g but center.K[k]=0 does not make sense\n\
*** (note that center.r0 was renamed to center.z0 from version 2.2g)",
                   k,center.r0[k])) }
      if (center.on&2) {
        if (SQR(center.r0))
          prt("atom-based force K=(%g %g %g) harmonic from r0=(%g %g %g)",
                                   VARG(center.K),             VARG(center.r0));
        else
          prt("atom-based harmonic force K=(%g %g %g) specified",
                                            VARG(center.K)); }

#ifdef SLAB
      loop (k,0,NCENTER) {
        center.nn[k]=center.n[k];
        if (k) center.nn[k]+=center.nn[k-1];
        center.on |= center.Kz[k]!=0;
        center.Kz2[k]=2*center.Kz[k];
        center.dz[k]=center.z1[k]-center.z0[k]; }
      if (center.on&1)
        loop (k,0,NCENTER) if (center.nn[k] && center.Kz[k])
          if (center.z1[k]) {
            prt_("slab step K=%g centered at z=L/2%+g, from z0=+-%g to z1=+-%g\n  for n in [%d,%d)",
                center.Kz[k],center.z[k],center.z0[k],center.z1[k],
                k==0?0:center.nn[k-1],center.nn[k]);
            center.Kz2[k]=center.Kz[k]/2*Sqr(center.dz[k]);
            prt_("  step height=%g K  width=%g A",center.Kz2[k],center.dz[k]);
            if (thermostat) prt_("  Boltzmann factor=%g",exp(-center.Kz2[k]/T));
            _n }
          else
            prt("z-force K=%g centered at z=L/2%+g, from z0=+-%g for n in [%d,%d)",
                center.Kz[k],center.z[k],center.z0[k],
                k==0?0:center.nn[k-1],center.nn[k]);
#endif
    }
#endif /*# FREEBC */

    if (center.cmn<0) center.cmn=0;

    if (center.cmn==0)
      if (center.cmK[0]!=0 || center.cmK[1]!=0 || center.cmK[2]!=0)
        prt("Nonzero center.cmK specified with center.cmn=0 ==>\n\
  center.cmn=%d (=No.N=all molecules) assigned",center.cmn=No.N);

    center.cmmass=0;
    if (center.cmn) {
      int n,i;

      loop (n,FROM,center.cmn) {
        molecule_t *mn=cfg+n;
        siteinfo_t *si=spec[mn->sp]->si;

        loop (i,0,mn->ns)
        center.cmmass+=si[i].mass; }

      prt("Whole-configuration harmonic force K*(r_CM-r0)^2 to center-of-mass specified\n\
  for %d molecules of total mass %g (prog.units):",center.cmn,center.cmmass);
      loop (i,0,DIM) if (center.cmK[i]!=0) {
        double tau=sqrt(center.cmmass/(2*center.cmK[i]));

        prt("K%c = %g  ==> tau = %g ps (oscillation period=%g ps)",i+'x',
            center.cmK[i],tau,2*PI*tau); } }

#ifndef FREEBC
    if (center.cmn || center.on)
      prt("WARNING: the above forces violate periodic b.c.\n");
#endif

//    VVO(Eext.E,=Eelst,/352259391.)
    /* conversion of V/m to internal prog. units */
    VVO(Eext.E,=Eelst,*(lengthunit*chargeunit/energyunit))

    Eext.isE=SQR(Eext.E)>0;
    if (Eext.E[0]!=0 && Eext.E[1]==0 && Eext.E[2]==0) Eext.isE=-3;
    if (Eext.E[1]!=0 && Eext.E[0]==0 && Eext.E[2]==0) Eext.isE=-2;
    if (Eext.E[2]!=0 && Eext.E[1]==0 && Eext.E[0]==0) Eext.isE=-1;
    if (Eext.isE) prt("WARNING: electric field has been specified\n\
* energy in ionic system may not be conserved");
    if (Eext.isE && lag.cond==0) prt("* lag.cond=0 so that the current density is not measured and conductivity\n\
  cannot be calculated");
#ifdef COULOMB
    if (Eext.isE<0) prt("M[%d] will be monitored instead of |M|",Eext.isE+3);
#endif

    if (No.c && (thermostat==T_ANDERSEN || thermostat==T_MAXWELL))
      WARNING(("Constraint dynamics and atom-based Maxwell/Andersen thermostat:\n\
*** The results will be likely severely affected.\n\
*** Consider center-of-mass-based Andersen and Maxwell thermostats instead\n\
*** (thermostat=\"AndersenCM\" or thermostat=\"MaxwellCM\")."))

    En.corr=cutcorr(LJcutoff); /* called because of counting rdf->npairs */

#ifdef WIDOM
    if ( (widom.n || widom.spreal>=0) && thermostat==0)
      WARNING(("Widom insertion or scaling requires a thermostat because\n\
*** temperature T is used to calculate the insertion probability"))
    widom.corr=Widomcutcorr(widom.spreal,widom.sp,LJcutoff);
    if (widom.spreal>=0 && widom.n)
      ERROR(("Widom insertion (widom.n>0) and Widom scaling (widom.spreal>=0) are exclusive"))
#endif

#if defined(TWODIM) || defined(FREEBC)
    En.corr=0; /* not supported */
#endif /*# defined(TWODIM) || defined(FREEBC) */

    prt("mass of whole configuration in prog.units: ");
    put2(No.mass,No.free_mass)
    prt("mass = %.8g g/mol = %.8g kg", No.mass*munit,No.free_mass*massunit);
    prt("rho = %.12g / V  (for rho in kg m^-3 and V in A^3)", rhounit*No.mass);
    _n

    if (option('f')) {
      /* autodiffusion and structure factor from stored cfgs */
      double d=dt.plb;

      if (option('f')<0) d=dt.cfg;
      diffSF(d,h*noint,no,init,loadreversed,nspec);
      goto TheEnd; }

    if (option('m')==1 && init>=0) init=-1;

    initgammas();

    if (init==0) {
      StaLoad(Fn("sta"));
      issta++;
      if (nomax>0) if (StaN("Tkin")>=nomax) {
        fprintf(stderr,"already nomax=%d cycles has finished - stop\n",nomax);
        prt("already nomax=%d cycles has finished - stop\n",nomax);
        exit(1); } }
    LRFree();

    if (init==0) loadrdf(rdf.grid);
#ifdef DIHHIST
    if (init==0) loaddih(dih.grid);
    if (dih.dcp) dihcp=fopen(Fn("dcp"),init<2?"at":"wt");
#endif /*# DIHHIST */
#ifdef HARMONICS
    if (init==0) loadharmonics(rdf.grid,rdf.cutoff);
#endif /*# HARMONICS */
#ifdef SLAB
    if (init==0) loaddpr(densprof.grid);
#endif /*# SLAB */


/* SIMNAME.000, ... to plb conversion, no simulation */
    if (init==999) {
      openplayback(option('y'),0);
      loopto (i,1,no) {
        loadcfg(i,loadreversed,0,&RvdW); /* 1st pass */
        loadcfg(i,loadreversed,0,NULL);  /* 2nd pass */
        if (simils.frommol) ERROR(("truncated %s - some molecules undefined",lastFn));
        writeplayback(); }
      return 0; }

    /* fix sites/axes by SHAKE constraint, measure forces/torques */
#ifdef ANCHOR
    initanchor(Fn("fix"),init,drmax);
#endif /*# ANCHOR */

    /*** initialize/load configuration ***/
    if (init>=3) {
#ifdef COULOMB
      if (init<10) initrspace();
#endif /*# COULOMB */
      if (init>=10) {
        readasc(0,1);
        if (init>=11) readasc(1,1);
        if (init>=12) readasc(2,1);
        if (init>12) ERROR(("init=%d not implemented",init)) }
      else if (init==5)
        initcryst((int)pins,Emax);
      else {
        if (MC && option('y')) {
          openplayback(option('y'),0);
          plbopened=1; }
        initcfg(pins,Emax,
                option('y')?dt.plb==0?1:(int)(dt.plb/h+0.5):0); } }
    else if (init>=0) {
      loadcfg(-1,loadreversed,option('['),NULL); /* 2nd pass */
      if (init==2) t=0;
      if (stop>0 && t>stop-h/2) sig=3;
      if (simils.frommol) {
#ifdef COULOMB
        /* imprecise for init */
        initerfc(el.grid,el.minqq,box.cutoff+ERFCPLUS+el.minqq,box.cutoff,option('v')&4?-el.alpha:el.alpha,1);
#endif
        initcfg(0,Emax,0);
        simils.frommol=0; }
      option('[')=0; }
    else if (init<0) {
      if (option('y')) ERROR(("Options -y (write plb-file) and init<0 (initialize cfg from a plb-file)\n\
*** are not allowed in one run.  Remedies:\n\
*** 1. use utility plb2cfg to convert SIMNAME.plb to SIMNAME.cfg\n\
*** 2. run cook -w -y0 with init<0 no=0 to convert SIMNAME.plb to SIMNAME.cfg\n\
*** Then run cook or cook -y### again with init=2"))
      readplayback(-init,loadreversed,1);
      prt("initialized from playback frame %d, continuing as with init=2",
        frame0=-init);
      init=2; }

    if (init<0 || init>1) a[0]->logs=0; /* for Nose: remove drift */
    // better not...  a[1]->logs=0;

    /* fix sites by springs, do not measure force */
    if (option('k')>0) {
      makefixa();
      loadfixa();
      savefixa(); }

    if (init>1) {
      En.tot=3e33; /* to ignore */
      a[0]->logs=0; }

    if (initvel) Maxwell(0,initvel,1);
    initvel=0; /* just once */

    if (option('h')) {
      int valence=abs(option('h'));

      while (valence) {
        prt("WARNING: %d %d-groups or neighbors centered\n",
            centergroups(a[0],valence%10), valence%10);
        valence/=10; }
      if (option('h')>0) option('h')=0; }

    /* init=0,1: apppend playback; init=2,3: new playback */
    if (option('y')) {
      if (!plbopened) openplayback(option('y'),init<2&&init>=0);
      if (init>=2) if (justnow(dt.plb,h/2) || dt.plb==0) writeplayback(); }

#ifdef TWODIM
    V=box.L[0]*box.L[1];
#else /*# TWODIM */
    V=box.L[0]*box.L[1]*box.L[2];
#endif /*# TWODIM */

#ifdef LINKCELL
    lcsetup(L);
#endif /*# LINKCELL */

#if PARALLEL
    initparallel();
#if PARALLEL==1 || PARALLEL==3
    parallelizerdf(1);
#endif
#endif

    addshift(shift);
    if (init>=2) normalize(-2);

    putv(L)
    putv(box.L)
    put2(box.cutoff,V)

    constrd.Pcorr=En.corr/(V*V)*Punit;
    underline("cutoff corrections");
    prt("Lennard-Jones cutoff=%g %s ==> cutoff correction for homogeneous fluid\n\
En.corr=%.9g K A^3 (Ecorr=En.corr, Pcorr=En.corr/V^2)\n\
prog.units: Ecorr=%g K  Pcorr=%g p.u.",
      fabs(LJcutoff),LJcutoff>0?"A":"sigma",En.corr,En.corr/V,En.corr/V/V);
    prt("Ecorr=%g %s, Pcorr=%g Pa",
        En.corr/V*Eunit/SIE,SI?"J/mol":"kcal/mol",constrd.Pcorr);
    prt("The above corrections are %sincluded in P and E (because corr=%d)",
      corr?"":"NOT ",corr);
    if (!corr) constrd.Pcorr=En.corr=0;
#ifdef SLAB
    if (densprof.grid)
      prt("\n\
>>> SLAB version hack: For compatibility, the above corrections are included\n\
>>> which is incorrect in slab geometry. Correct corrections will be calculated\n\
>>> later from the density profile (you need to set densprof.slab&1).");
#endif /*# SLAB */

    halftcyc=h*noint/2;

#ifdef COULOMB
    underline("initialization of electrostatic forces");
    if (Ewaldtest(L)) goto TheEnd;
    prt("Ewald set final values: cutoff=%g el.alpha=%g el.kappa=%g",
                             box.cutoff,   el.alpha,   el.kappa);
#endif /*# COULOMB */

    if (quit<0) quit=debugcfg(quit);

#ifdef MARK
    mark_do(no,loadreversed);
#else /* everything suppressed for !MARK */ /*# MARK */

    loadreversed=0; /* where? */

    if (init%10>=3) {
      /* thus, forces are not calculated for init=10 */
#ifdef Lagrange
      measure=1; /* debugging only! */
      /* necessary rhs if Verlet? */
#ifdef POLAR
      { /* OOPS! omegap may be -9 here to be set later.... */
        double o=omegap;
        omegap=0.95;
        rhs(a[2],a[0],a[1]);
        omegap=o; }
#else
      rhs(a[2],a[0],a[1]);
#endif
      rescalecfg(a[2],RESCALE_XYZ,h*h/2,0,0); /* for Gear only? */

#ifdef POLAR
      /*.....  memset(polarrof(cfg,a[2]->rp),0,No.s*sizeof(vector));*/
      loop (i,1,GearOrder)
        copy(polarrof(cfg,a[i]->rp),polarrof(cfg,a[0]->rp),No.s*sizeof(vector));
#endif /*# POLAR */

#else /* Hamilton */ /*# Lagrange */
      rhs(a[1],a[0]); hrescale(a[1],0,h); hrescale(a[1],1,h);
#endif /*# Lagrange */
      }

#if 0
#include "foolproo.c"
#endif /*# 0 */

    /* columns 4,5 of the convergence profile */
    if (tau.P==0 && tau.rho==0 && tau.R==0)
      initCP(no,CPnbit,dV==0?(No.first?"Ep0":"Ein"):constrd.PdVname,"P",corr);
    else
      initCP(no,CPnbit,tau.R?"RvdW":"rho",dV==0?"P":constrd.PdVname,corr);

    if (tau.R==0) {
      if (RvdW!=0) prt("RvdW=%g in cfg-file found but ignored (ble-file value applies)\n\
RvdW=0 was set for a subsequent write to the cfg-file\n\
so that next time tau.R is specified, RvdW from the ble-file will be used",RvdW);
      RvdW=0; }
    else
      if (RvdW==0) {
        RvdW=sitedef[tau.i].LJ[0].RvdW;
        prt("RvdW adjustment, tau.i=%d: RvdW=0 (loaded or initialized) replaced by force field value %g",tau.i,RvdW); }
      else {
        prt("RvdW adjustment, tau.i=%d: RvdW=%g loaded from %s",tau.i,RvdW,Fn("cfg"));
        sitedef[tau.i].LJ[0].RvdW=RvdW; }

#ifdef DIHHIST
    gauchetrans(-3);
#endif /*# DIHHIST */

    inittime(option('t')?8:0);
    time(&starttime);
    nsteps=0;

#ifdef POLAR
    put3(epsp,maxitp,omegap)
    epsp0=epsp;
    if (epspq>=1) ERROR(("epspq=%g is invalid",epspq))
    if (epspq>0) epsp *= 1-epspq;
    if (epspx==0) epspx=epsp*.1;
    if (epsp0==epspx) epspx*=.9999999999;
    put3(epsp,epspq,epspx)
#endif /*# POLAR */

    {
      char *st=ctime(&starttime);

      if (st[24]=='\n') st[24]=0;
      fprintf(stderr,"%d cycle%s started at %s\n",no,"s"+(no==1),st);
      prt("\n*** %d cycle%s started at %s",no,"s"+(no==1),st);
    }

    if (option('b')<0) {
      /* for batch mode -s0 only - default -b */
      option('b')=100000*abs(option('b'))/(No.s*noint)+1; }

    fflush(out); /* added 10/00 */

    if ((drift&DRIFT_WHEN)==DRIFT_START) {
      drift&=DRIFT_DEPEND;
      prt("no CM/v/angular drift corrected anymore because drift&%d specified\n\
final drift = %d = %s",DRIFT_START,drift,int2sumbin(drift)); }

    /* <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
    for (icyc=0; icyc<no && sig==0; icyc++) {

#ifdef DISTURB
      Disturb(0,disturb[0]);
      Disturb(1,disturb[1]);
#endif /*# DISTURB */

#ifdef DIHF
      /* WARNING this is something very obsolete... */
      /* warning : memset for doubles generally not portable */
      memset(dihf,0,sizeof(dihf));
      memset(dihfcov,0,sizeof(dihfcov));
#endif /*# DIHF */

      lastEtot=En.tot;

      if (option('d')==2) distancecheck();

      if (option('m')==1) {
        /* read playback instead of simulating */
        readplayback(icyc+frame0,loadreversed,1);
        prt("frame %d/%d read",icyc+frame0,no);
        icyc+=noint-1; }

      else loopto (iint,1,noint) {

        if (option('s') && option('b')>3)
          fprintf(stderr,"%d/%d %d/%d t=%f   \r",iint,noint,icyc,no,t);

        /* friction-like isobaric ensemble: Pscale kept during cycle */
        if (tau.P>0)
          V=rescalecfg(a[0],rescale|RESCALE_L,Pscale[0],Pscale[1],Pscale[2]);

        /* adjusting an R-like parameter */
        if (tau.R>0) {
          sitedef[tau.i].LJ[0].RvdW/=Pscale[0];
          setss(&sstab[tau.i][tau.i],tau.i,tau.i,LJcutoff,0); }

        if (option('d')>2) distancecheck();

        if (sig==5) break;

        measure=iint==noint;
#ifdef SHAKE
        if (shake) Shake(epsc,
          thermostat==T_MAXWELL||thermostat==T_MAXWELL_CM ? (double)justnow(tau.T,h/2) :
          thermostat==T_ANDERSEN||thermostat==T_ANDERSEN_CM ? h/tau.T : 0);
        else
#endif /*# SHAKE */
          {
#ifdef POLAR
            Gear2pol(No.eq,No.s*DIM,(option('p')/10)%10,(option('p')/100)%10,a);
#else /*# POLAR */
            Gear(No.eq,a);
#endif /*# POLAR */
            CPUtime("Gear"); }

        nsteps++;

        if (!shake) {
          int n;
          vector *rp=a[0]->rp;
          vector *rpvel=a[1]->rp;

          if (thermostat==T_ANDERSEN) Maxwell(-1,No.N,h/tau.T);
          if (thermostat==T_ANDERSEN_CM) MaxwellCM(h/tau.T);
          if (thermostat==T_MAXWELL && justnow(tau.T,h/2)) Maxwell(-1,No.N,1);
          if (thermostat==T_MAXWELL_CM && justnow(tau.T,h/2)) MaxwellCM(1);

          if (measure && option('c')&4) {
            En.r2=constrainterror(a[0],a[1]);
#ifdef Lagrange
            En.v2=vconstrainterror;
#endif /*# Lagrange */
            }

          if (!(option('c')&1)) {
            loop (n,FROM,No.N) if (cfg[n].nc)
              constrit[cfg[n].sp].nit[COR_BOTH] +=
                Scorrect(cfg+n,rof(cfg+n,rp),rpvel,epsc); }
          else
            Lcorrect(a[0],a[1]);
          CPUtime("corr"); }

        normalize(-1);

        if (measure && option('c')&8) {
          En.r3=constrainterror(a[0],a[1]);
#ifdef Lagrange
          En.v3=vconstrainterror;
#endif  /*# Lagrange */
          }

        CPUtime("norm");

        if (option('y')) if (justnow(dt.plb,h/2)) writeplayback();

#ifdef POLAR
        if (epspq>0) epsp = epsp*epspq + (1-epspq)*epsp0;
#endif /*# POLAR */

#ifdef __EMX__
        if (_read_kbd(0,0,0)==27) {
          sig=option('i');
          fprintf(stderr,"ESC hit, action = %d = %s\n",sig,action(sig)); }
        fprintf(stderr,"%d:%d/%d:%d  \r",iint,icyc+1,noint,no);
#endif /*# __EMX__ */

        if (No.c) doconstrit(omegac);
        if ((drift&DRIFT_WHEN)==DRIFT_STEP) removedrifts(option('v')&64);

        init_append|=1;
        } /* iint */

      if (!finite(En.tot)) {
        sig=2;
        WARNING(("En.tot is infinite or not a number => stop as if Ctrl-C")) }

      if (No.c) { /* constraint errors */
        int i,sp;
        /* constraint iterations reported every cycle even if known every step */
        /* ... not for SHAKE */
        loop (i,0,COR_LAST) loop (sp,0,nspec) if (constrit[sp].nit[i])
          StaAdd(string("constrit%d spec%d",i,sp),
            (double)constrit[sp].nit[i]/spec[sp]->N);
#ifdef Lagrange
        if (En.v1) StaAdd("v constr err before",En.v1);
        if (En.v2) StaAdd("v constr err mid",En.v2);
        if (En.v3) StaAdd("v constr err after",En.v3);
#else /*# Lagrange */
        StaAdd("gsq",gsq);
#endif /*# Lagrange */
        if (En.r1) StaAdd("constr err before",En.r1);
        if (En.r2) StaAdd("constr err mid",En.r2);
        if (En.r3) StaAdd("constr err after",En.r3); }

      if (drift&DRIFT_DEPEND) depend_rv(a[0],a[1]); /* new in 2.0b */

      /* there is a theoretical problem which value is `better pressure'... */
/*.....  Pvir=(((DIM*(No.s-No.depend)-No.c)*En.T+En.vir)/DIM+corr/V)/V;*/

      V=box.L[0]*box.L[1]*box.L[2];
#ifdef FREEBC
      V=9e9; // UGLY!
#endif
      En.P=((En.kin*2+En.vir)/DIM+En.corr/V)/V*Punit;

      if (nomax>0) {
        i = issta ? StaN("Tkin") : 0;
        if (i+1>=nomax) sig=3; }

      if (stop>0)
        if (t>stop-h/2) sig=3;

      if (option('w')>1) expectedbuffer+=16;
      if (expectedbuffer>7500) {
        expectedbuffer=0;
        waitfordiskspace(64);
        fflush(out); }

/*.....  if (!option('g') || out!=stdout)*/
      if (icyc==0 || icyc+1==no
          || sig
          || justnow(dt.prt,halftcyc))

        if ( (option('v')&1) && option('m')!=1) {

          if (option('w')>1) expectedbuffer+=88;

#ifdef POLAR
#ifdef REDUNITS
          if (icyc==0)
            header("    t         T       Epot       Etot      rho      P     polerr  polit  cerr1 ");
          /* never active...*/
          prt("%8.3f %7.4f %9.2f %12.4f %7.4f %8.3f%10.2e%10.2e",
              t,    En.T,En.pot,En.tot,No.mass/V,
              En.P,En.polstderr,En.polnit,En.r1);
#else
          if (icyc==0)
            if (tau.R) header("   t/ps    Tkin/K   Epot/K    Etot/K     RvdE/A  P/MPa  polerr polit  cerr1 ");
            else header("   t/ps    Tkin/K   Epot/K    Etot/K  rho/kgm^-3 P/MPa  polerr polit  cerr1 ");
          if (tau.R) prt("%9.3f %6.1f %9.0f %12.2f %7.4f %8.3f%9.2e%2d%10.2e",
              t,    En.T,En.pot,En.tot,sitedef[tau.i].LJ[0].RvdW,
              En.P/1e6,En.polstderr,En.polnit,En.r1);
          else prt("%9.3f %6.1f %9.0f %12.2f %7.2f %8.3f%9.2e%2d%10.2e",
              t,    En.T,En.pot,En.tot,No.mass/V*rhounit,
              En.P/1e6,En.polstderr,En.polnit,En.r1);
#endif
#else
          double pcer=En.r1+En.r2+En.r3;

          if (pcer>0) pcer=-log10(pcer);
          else pcer=9.99;
          if (icyc==0) {
            prt("\nNOTE: pcer = -log10(summary constraint error), recommended pcer>6");
            if (T) prt_("NOTE: the graph of Tkin is in the range of [0,2*T]=[0,%g] K",T*2);
            else prt_("NOTE: the graph of Tkin is in the range of [0,0.2] K");
#ifdef REDUNITS
            header("    t         T       Epot       Etot      rho      P  pcer0-------Tkin------2T");
#else /*# REDUNITS */
            if (tau.R) header("   t/ps    Tkin/K   Epot/K    Etot/K     RvdE/A  P/MPa pcer0-------Tkin------2T");
            else header("   t/ps    Tkin/K   Epot/K    Etot/K  rho/kgm^-3 P/MPa pcer0-------Tkin------2T");
#endif /*# REDUNITS */
          }

#ifdef REDUNITS
          prt_("%8.3f %7.4f %9.2f %12.4f %7.4f %6.3f%5.2f",
              t,    En.T,En.pot,En.tot,No.mass/V*rhounit,En.P,pcer);
#else
          /*** standard nonpolar prt ***/
          if (tau.R) prt_("%9.3f %6.1f %9.0f %12.2f %7.4f %6.1f%5.2f",
               t,    En.T,En.pot,En.tot,sitedef[tau.i].LJ[0].RvdW,
               En.P/1e6,pcer);
          else prt_("%9.3f %6.1f %9.0f %12.2f %7.2f %6.1f%5.2f",
               t,    En.T,En.pot,En.tot,No.mass/V*rhounit,
               En.P/1e6,pcer);
#endif
          graph(En.T/(T+0.1*(T==0))*0.5,20);

#ifdef Lagrange
#ifdef SHAKE
          if (!shake)
#endif /*# SHAKE */
            if (En.r1>=0 || En.r2>=0 || En.r3>=0) prt(
             "                                                 %10.2e%10.2e%10.2e",
             En.v1,En.v2,En.v3);
#endif /*# Lagrange */

#ifdef erminmax
#ifdef SHAKE
          if (!shake)
#endif /*# SHAKE */
            prt("%8.5f %8.5f",sqrt(Erfc.min),sqrt(Erfc.max));
          Erfc.min=3e33; Erfc.max=0;
#endif /*# erminmax */
#endif
          }

#ifndef FREEBC
        {
          double minLh=fmin(fmin(box.Lh[0],box.Lh[1]),box.Lh[2]),perc;
          static double oldperc;
          static int info=1;

          if (box.cutoff>minLh*1.005) {
            perc=(box.cutoff/minLh-1)*100;
            if (info)
              prt("\
Explanation: The box size is variable in NPT simulation while the cutoff is\n\
  constant. If the cutoff is inappropriately set (e.g., to minimum L/2 at\n\
  simulation start which is the default), it may happen that the cutoff\n\
  exceeds minL/2. Overshooting by less than 0.5 %% is acceped without warning.\n\
  More than a few %% is inaccurate and not acceptable in productive runs");
            if (perc>oldperc)
              WARNING(("cutoff=%f exceeding minL/2=%f by %.2f%%\n\
*** (more warnings for cutoff exceeding minL/2 by more than %.2f%%)",
                        box.cutoff,minLh,perc,oldperc=perc*1.2))
            info=0; }
         }
#endif /*# FREEBC */

      if (sig==5) break; /* 2nd attempt */

      { /*** statistics and convergence profile ***/
#ifdef FREEBC
        double Epot=En.pot*Eunit;
#else
        double Epot=(En.pot+En.corr/V)*Eunit;
#endif
#ifdef RGYR
        double endtoend,Rgyr;
#endif /*# RGYR */
#ifdef WIDOM
        double Wid;
#endif /*# WIDOM */
#ifdef XSECTION
        static double Xsec;

        if (xs.freq && icyc%xs.freq==0) {
          StaSet(DT*abs(xs.freq),lag.err,2,lag.n);
          Xsec=Xsection(xs.grid,xs.mode,xs.RvdW,xs.Rscale,
                        xs.freq<0,xs.cp,xs.ns,xs.sizelimit,&xs.maxsize);
          StaSet(DT,lag.err,2,lag.n); }

#endif /*# XSECTION */

#ifdef RGYR
        Rgyr=measureplus(&endtoend,rg.end,rg.cp);
#else /*# RGYR */
        measureplus();
#endif /*# RGYR */

        bounddr(drmax);
        if ((drift&DRIFT_WHEN)==DRIFT_CYCLE) removedrifts(option('v')&64);

        if (option('m')!=2) {
          /*** independent pressure measurement via <dU/dV> */
          /* SHAKE: ==> moved to constrd.c, shake() */
          measureP(1);
          measureP(2); }

        ssdistance();

/*** WARNING: watch the former place of this part*/
#ifdef WIDOM
        if (widom.spreal>=0) Wid=XWidom(widom.spreal,widom.sp);
#ifdef SLAB
        else Wid=Widom(widom.sp,widom.n,widom.z0,widom.z1,widom.dz,widom.mode);
#else /*# SLAB */
        else Wid=Widom(widom.sp,widom.n,widom.corr);
#endif /*# SLAB */
#endif /*# WIDOM */

        if (CPbuf) {
          float *CPrec=CPbuf+icyc*NCP;
          int count=0;

          memset(CPrec,0,NCP*sizeof(CPrec[0]));
#define recordCP(X) { if (count<NCP) CPrec[count++]=(float)(X); }
          /* order of recordCP must match CPinfo in simmeas.c */
          /*column (index is by 1 less)*/
          /*1*/ recordCP(En.tot)
          /*2*/ recordCP(En.T)
          /*3*/ recordCP(Epot/SIE)

          /* columns 4,5 of the convergence profile */
          if (tau.P==0 && tau.R==0 && tau.rho==0) {
            if (dV==0) {
              if (No.first) recordCP(En.pot0*Eunit/SIE)
              else recordCP(En.intra*Eunit/SIE) }
            else
              recordCP(En.PdV)
            recordCP(En.P) }
          else {
            if (tau.R) recordCP(sitedef[tau.i].LJ[0].RvdW)
            else recordCP(rhounit*No.free_mass/V);
            if (dV==0) recordCP(En.P)
            else recordCP(En.PdV) }
          /*6*/ recordCP(En.T_in)
          /*7*/ recordCP(En.T_tr)
#ifdef COULOMB
          /*8*/ if (Q) {
                  if (Eext.isE<0) recordCP(Q->M[Eext.isE+3])
                  else recordCP(sqrt(SQR(Q->M))) }
                else count++;
#endif /*# COULOMB */
#ifdef SHEAR
          /*9*/ recordCP(En.Cv)
#endif /*# SHEAR */
#ifdef POLAR
          /*10*/ recordCP(En.self)
#endif /*# POLAR */
#ifdef DIHHIST
          /*11*/ recordCP(gauchetrans(dih.cp))
          if (dihcp) fprintf(dihcp,"\n");
#endif /*# DIHHIST */
#ifdef RGYR
          /*12*/ recordCP(Rgyr)
          /*13*/ recordCP(endtoend)
#endif /*# RGYR */
#ifdef XSECTION
          /*14*/ recordCP(Xsec) /* warning: need not be at the same frequency! */
#endif /*# XSECTION */
#ifdef WIDOM
          /*15*/recordCP(Wid)
#endif

          En.logs=a[0]->logs;
          En.dlogs=a[1]->logs;
          if (ssd0) {
            /* variables recorded in SIMNAME.cpi (incl. distances) */
            struct ssd_s *ssd;

            for (ssd=ssd0; ssd; ssd=ssd->next)
              if (ssd->indx[0]==-1) recordCP(*ssd->u.q*Eunit/SIE)
              else if (ssd->indx[0]==-2) recordCP(*ssd->u.q)
              else recordCP(ssd->u.dist); } }

        StaSet(DT,lag.err,2,lag.n);
        StaAdd("Tkin",En.T); issta=1;
        StaAdd("Tin",En.T_in);
        StaAdd("Ttr",En.T_tr);
        StaAddE("Epot",Epot);
        StaAddE("Ein",En.intra*Eunit);
        if (No.first)
          StaAddE(string("Epot0 (<%d)",No.first),En.pot0*Eunit);
#ifdef COULOMB
        if (Q) {
          StaAdd("M[0]",Q->M[0]);
          StaAdd("M[1]",Q->M[1]);
          StaAdd("M[2]",Q->M[2]);
          StaAdd("M^2",SQR(Q->M)); }
#endif   /*# COULOMB */
#ifdef Lagrange
        StaSet(DT,2,2,0);
        StaAdd("virial",En.vir);
        StaSet(DT,lag.err,2,lag.n);
        StaAdd("P [Pa]",En.P);
        if (dV) StaAdd("P-PdV [Pa]",En.P-En.PdV);
        if (tau.P) StaAdd("V",V);
        if (tau.R) StaAdd("RvdW",sitedef[tau.i].LJ[0].RvdW);
#endif /*# Lagrange */

#if PRESSURETENSOR&PT_OFF
        if (lag.visc) {
          StaSet(DT,lag.visc,2,0);
          StaAdd("Ptyz [Pa]",En.Ptens[3]);
          StaAdd("Ptzx [Pa]",En.Ptens[4]);
          StaAdd("Ptxy [Pa]",En.Ptens[5]); }
#endif /*# PRESSURETENSOR&PT_OFF */

        StaSet(DT,lag.err,2,lag.n);
        StaAdd("Etot [K]",En.tot);
        if ((thermostat==1 || thermostat>2) && tau.E==0)
          LRAdd("Etot",1,t,En.tot);
        if (lastEtot<2.999e33) {
          lastEtot -= En.tot;
          StaSet(DT,2,2,0);
          StaAdd("|dEtot|",fabs(lastEtot));
          StaAdd("dEtot^2",lastEtot*lastEtot); }
#if (PRESSURETENSOR&PT_ANY) == PT_ANY
        StaAdd("tr(Ptens)/3+corr [Pa]",SUM(En.Ptens)/3+En.corr/V/V*Punit);
#endif

#ifdef DIHHIST
        StaSet(DT,lag.err,2,lag.n);
        gauchetrans(-2);
#endif /*# DIHHIST */

        if (ssd0) {
          struct ssd_s *ssd;

          StaSet(DT,lag.err,2,lag.n);
          for (ssd=ssd0; ssd; ssd=ssd->next) if (ssd->stat) {
            if (ssd->indx[0]==-1)
              StaAddE(ssd->name,*ssd->u.q*Eunit);
            else if (ssd->indx[0]==-2)
              StaAdd(ssd->name,*ssd->u.q);
            else
              StaAdd(ssd->name,ssd->u.dist); } }

#ifdef HARMONICS
        measureharmonics();
#endif /*# HARMONICS */

#ifdef SLAB
        measuredpr();
#endif /*# SLAB */
        } /*** statistics+CP ***/

#ifdef DIHF
      /* WARNING this is something very obsolete... */
      /* Fourier analysis of the dihedral angle distribution */
      {
        int type,k;

        loop (type,0,MAXDIH) {
          static char st[12];

          StaSet(DT,lag.err,2,lag.n);
          loopto (k,0,DIHF) {
            if (k==2) StaSet(DT,2,2,0);
            sprintf(st,"dih%i*cos%i",type,k);
            StaAdd(st,dihf[type][k]); }
          sprintf(st,"dih%i cov",type);
          StaAdd(st,dihfcov[type]); } }
#endif /*# DIHF */

      if (dt.cfg!=0) {
        int n=(t+halftcyc)/dt.cfg;

        if ( n > (int)((t-halftcyc)/dt.cfg) ) {
          time(&stoptime);
          if (tau.R) RvdW=sitedef[tau.i].LJ[0].RvdW;
          savecfg(n,stoptime,&RvdW); } }

      if (tau.E!=0) {
        double Etot=En.tot;
        if (tau.P) Etot+=P/Punit*V; /* enthalpy in K */
#ifdef Lagrange
        rescalecfg(a[1],RESCALE_XYZ,exp(halftcyc*(E-En.tot)/En.kin/tau.E),0,0);
#else /*# Lagrange */
        hrescale(a[0],1,exp(halftcyc*(E-En.tot)/En.kin/tau.E));
#endif /*# Lagrange */
        }
      StaSet(DT,2,2,0);
      if (drift&7) StaAdd("CM drift",En.cmdrift);
      if (drift&7*8) StaAdd("v drift",En.vdrift);
      if (drift&7*64) StaAdd("angular drift",En.angdrift);
#ifdef FREEBC
      if (tau.rho || tau.P || dV) ERROR(("tau.rho, tau.P, dV not supported with FREEBC"))
#else /*# FREEBC */

      /* rescale to reach given density or L */
      if (tau.rho) {
        if (rhorescale(rescale,tau.rho,L,halftcyc)) {
          if (rhoreached) prt("^^^^^^^^^^^^^^^^ the reference box (density) has been reached ^^^^^^^^^^^^^^^^");
          rhoreached=0; } }

      /* friction-like (Berendsen) barostat or adjusting molecule size */

//prt("tr(P)=%g Pvir=%g",SUM(En.Ptens),En.P-En.corr/V/V*Punit);
//prt("tr(Pvir)=%g",SUM(En.Pvir));
//prt("tr(Pkin)=%g Ekin=%g",SUM(En.Pkin)*V/Punit*1.5,En.kin);

      if ((rescale & RESCALE_PT) && tau.P!=0) {
#if (PRESSURETENSOR&PT_ANY) == PT_ANY
        int ii;
        double beta_T=V/(No.f*En.T); /* compressibility */

        if (bulkmodulus) beta_T=Punit/3/bulkmodulus;

        if (dV)
          ERROR(("implementation limitation:\n\
*** pressure tensor components calculated by virtual volume (area, shape)\n\
*** change (with dV set) cannot be used for shape change (rescale & %d)",RESCALE_PT))

        loop (ii,0,DIM)
          Pscale[ii]=scaling(tau.P,noint,
            h*beta_T*(P-En.Ptens[ii]-En.corr/Sqr(V)*Punit)/Punit,
            maxscale);
/*  Pscale=exp((tau.P+tau.R<0?noint:-1)*h*V/No.f/En.T/(tau.P+tau.R)*(P/Punit-locP)); */

        if (rescale & RESCALE_XisY) Pscale[0]=Pscale[1]=sqrt(Pscale[0]*Pscale[1]);

        if (tau.P<0) /* rescaling */
          rescalecfg(a[0],rescale|RESCALE_L,Pscale[0],Pscale[1],Pscale[2]);
#else
        ERROR(("rescale & RESCALE_PT (%d) was specified but this version\n\
*** of cook was not compiled with needed #define PRESSURETENSOR 3 (or more)",RESCALE_PT))
#endif
        }

      if ( (tau.P!=0 || tau.R!=0) && !(rescale & RESCALE_PT)) {
        double Ploc;
        double tauloc=tau.P;
        /* tau<0: rescaled every cycle
           tau>0: rescaled every step (keeps scaling factor through whole cycle)
           note: P and Pvir are in Pa */
        double beta_T=V/(No.f*En.T); /* compressibility */

        if (bulkmodulus) beta_T=Punit/3/bulkmodulus;

        if (tau.R) tauloc=tau.R;
        if (dV) Ploc=En.PdV; else Ploc=En.P;

        Pscale[0]=Pscale[1]=Pscale[2]=scaling(tauloc,noint,h*beta_T*(P-Ploc)/Punit,maxscale);

        if (tau.P<0) /* rescaling */
          rescalecfg(a[0],rescale|RESCALE_L,Pscale[0],Pscale[1],Pscale[2]);
        if (tau.R<0) {
          /* adjusting an R-like parameter */
          sitedef[tau.i].LJ[0].RvdW/=Pscale[0];
          setss(&sstab[tau.i][tau.i],tau.i,tau.i,LJcutoff,0); } }
#endif

      V=L[0]*L[1]*L[2];

      init_append|=2;

      if (!option('s') && option('b')) if (icyc%-option('b')==0) {
        /* raises sig if SIMNAME.stp exists */
        FILE *stp=fopen(Fn("stp"),"rt");

        if (stp) {
          fprintf(stderr,"%s exists, interrupting\n",lastFn);
          sig=4;
          fclose(stp); } }

      } /*** icyc (# of cycles=no, or until ^C) ***/
      /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> */


    if (option('v')&1) header("");
    if (option('d')) distancecheck();

#ifdef POLAR
    if (epspq>0) {
      put3(epspq,epsp,epsp0)
      epsp=epsp0; epspq=-fabs(epspq); }
#endif /*# POLAR */

#if PARALLEL
    printpartimes();
#endif /*# PARALLEL */

/*.....    if (option('g')) { show(1); show(0); }*/

    time(&stoptime);
    stop_start=stoptime-starttime;

    prt_("\n*** main cycle stopped at %lu = %s",stoptime,ctime(&stoptime));
    prt("duration %ld s = %.2f m (%g ps = %d steps of %d*%d)",
        stop_start,stop_start/60.0, nsteps*h, nsteps,icyc,noint);
    if (nsteps && stop_start>0)
      prt("speed = %g cycles/min = %g cycles/hour = %g ps/day",
          60.0/((double)stop_start/nsteps*noint),
          3600.0/((double)stop_start/nsteps*noint),
          86400.0*h/((double)stop_start/nsteps));
#ifdef XSECTION
    put(xs.maxsize)
#endif /*# XSECTION */

    if (option('t')) {
#ifdef CHEAPTIME
      if (nsteps) {
        int i=stop_start%nsteps;

        if (i>nsteps/2) i=nsteps-i;
        if (i<=(int)sqrt(nsteps)) prt("* possible interference *"); }
#endif
      printtime(); }

    _n

    if (option('s') && option('b')>1) fputc('\a',stderr);
    fprintf(stderr,"%d of %d*%d steps  t=%.3fps  in %lds=%.2fm  stop at %s",
            nsteps,icyc,noint,t,
            stop_start,stop_start/60.0,ctime(&stoptime));

    underline("statistics");
    if (option('v')&2) StaPrintAll(NULL);
    if (option('v')&4) StaPrintAll("");
    else prt("HINT: to get detailed statistics (as prior V2.4a), use `staprt %s'",simname);
    _n

#ifdef WIDOM
#include "widomres.c"
#endif

#ifdef DIHHIST
    printdihhist();
#endif /*# DIHHIST */

    if (tau.P) prt("rho = mass/<V> = %g",rhounit*No.free_mass/StaMean("V"));

#ifndef REDUNITS
    prt("<Epot>/molecule   = %11.5f kcal/mol = %g K",
        StaMean("Epot [*")/No.N/kcal*SIE,
        StaMean("Epot [*")/No.N/Eunit*SIE);
    prt("<Eintra>/molecule = %11.5f kcal/mol = %g K",
        StaMean("Ein [*")/No.N/kcal*SIE,
        StaMean("Ein [*")/No.N/Eunit*SIE);
#endif /*# REDUNITS */

#if 1 /* mean quadratic dipolemoments */
    {
      int n;
      double D;
      char s[16];

      loop (n,0,nspec) if (spec[n]->N) {
        sprintf(s,"dipmom[%d]^2",n);
        if ( (D=StaMean(s)) ) {
          D=sqrt(D);
          prt("sqrt(%s) = %f = %.4f D",s,D,D*Debye); } } }
#endif /*# 1 */

#if (PRESSURETENSOR&PT_ANY) == PT_ANY
    { /*
         this actually duplicates Pvirerr, but is more fool-proof
         because is based on the final (only uncorrected) pressure tensor
      */
      double x=StaMean("P [Pa]")-StaMean("tr(Ptens)/3+corr [Pa]");

      prt("\n%sressure tensor accuracy: P-tr(Ptens)/3=%g Pa",
          fabs(x)>1e6?"WARNING: Low p":"P",x);
      if (fabs(x)>1e7) WARNING(("Too low pressure tensor accuracy (%g Pa)",x))
    }
#endif /*# PRESSURETENSOR&PT_ANY */

#if PRESSURETENSOR&PT_MOM
    {
      char x;
      underline("kurtosis and normalized 6-th central moment of v_CM");
      header("DIR  kurtosis 6-th moment ");
      loop (x,'x','x'+DIM) {
        double E1=StaMean(string("PKin.%c%c [Pa]",x,x))/No.N;
        double E2=StaMean(string("PKin2.%c%c [Pa^2]",x,x))/No.N;
        double E3=StaMean(string("PKin3.%c%c [Pa^3]",x,x))/No.N;

        prt(" %c  %9.6f  %9.6f",x,E2/Sqr(E1)-3,E3/Cub(E1)-15*E2/Sqr(E1)+30); }
      header("");
    }
#endif
    if (tau.P) V=StaMean("V");
    /*** dielectric constant ***/
#ifdef COULOMB
    if (Eext.isE==0) {
      if (Q) prtdielconst(V,1);
      else prtdielconst(V,0); }
    else
      prt("EXTERNAL ELECTRIC FIELD (%g %g %g) V/m:\n\
* in ionic systems the conductivity in S/m is Jx/Ex or Jy/Ey or Jz/Ez,\n\
  where the current density J is provided in A/m2 (if lag.cond specified)\n\
* see the manual for determinimg the dielectric constant in dipolar systems",
          VARG(Eelst));
#endif /*# COULOMB */

    if (lag.cond) {
      underline("EMD conductivity hints");
      prt("(applies only in directions where Eelst_x,y,z=0)\n\
* calculate the time autocovariances of el. current density by commands:\n\
    staprt -t2 -nJx* %s\n\
    staprt -t2 -nJy* %s\n\
    staprt -t2 -nJz* %s\n\
* check that the data have sufficiently fine grid (otherwise decrease noint)\n\
* check that the data have sufficiently long lag (otherwise increase lag.cond)\n\
* calculate their running integrals by commands:\n\
    runint -e -m2 < %s.Jx.cov > %s.Jx.cov\n\
    runint -e -m2 < %s.Jy.cov > %s.Jy.cov\n\
    runint -e -m2 < %s.Jz.cov > %s.Jz.cov\n\
* plot these curves and determine the limiting averaged value:\n\
    plot %s.J?.run\n\
* the conductivity is this value multiplied by V/kT=%g",
          simname,simname,simname,
          simname,simname,simname, simname,simname,simname,
          simname,
          V*Cub(lengthunit)/energyunit/StaMean("Tkin")); }

#if PRESSURETENSOR&PT_OFF
    if (lag.visc) {
      underline("EMD viscosity hints");
      prt("(applies to xy only if the shear is also set)\n\
* calculate the time autocovariances of the off-diagonal components of the\n\
  pressure tensor (in Pa^2):\n\
    staprt -t2 -nPtxy* %s\n\
    staprt -t2 -nPtyz* %s\n\
    staprt -t2 -nPtzx* %s\n\
* check that the data have sufficiently fine grid (otherwise decrease noint)\n\
  and long enough lag (otherwise increase lag.visc):\n\
    plot %s.Pt??.cov\n\
* calculate their running integrals by commands:\n\
    runint -e -m2 < %s.Ptxy.cov > %s.Ptxy.run\n\
    runint -e -m2 < %s.Ptyz.cov > %s.Ptyz.run\n\
    runint -e -m2 < %s.Ptzx.cov > %s.Ptzx.run\n\
* plot these curves and determine the limiting averaged value:\n\
    plot %s.Pt??.run\n\
* the viscosity in Pa.s is this value multiplied by V/kT=%g",
          simname,simname,simname,
          simname,
          simname,simname,simname, simname,simname,simname,
          simname,
          V*Cub(lengthunit)/energyunit/StaMean("Tkin")); }
#endif

#ifdef SHEAR
    /*** shear viscosity ***/
    prtviscosity();
#endif /*# SHEAR */

    LRPrint(' '); _n

#ifdef ANCHOR
    if (anchor.f) fclose(anchor.f); anchor.f=NULL;
#endif /*# ANCHOR */

    /*** print pseudographs of rdf's ***/
    if (option('v')&1024) rdfgraph();

    sortmolecules(sort);
    sort=0;

    if (option('w')>1)
      waitfordiskspace(No.s
#ifdef POLAR
                           *2
#endif
                             /7+2*option('l')+no*NCP/200+256
#ifdef SLAB
                                                             +128
#endif
                                                                  );

    /*** save cfg, rdf and close cp files ***/
    if (option('w')) {
      if (tau.R) RvdW=sitedef[tau.i].LJ[0].RvdW;
      savecfg(-1,stoptime,&RvdW); }

    writeasc(); /* according to option -l */

#ifdef DIHHIST
    if (dihcp) fclose(dihcp);
#endif /*# DIHHIST */

    if (icyc || option('w')==1) {
#ifdef DIHHIST
      savedih(dih.grid);
#endif /*# DIHHIST */
#ifdef HARMONICS
      saveharmonics();
#endif /*# HARMONICS */
#ifdef SLAB
      savedpr(densprof.grid);
      printdpr(densprof.slab,dV,constrd.PdVname); /* density profiles and surface tension */
#endif /*# SLAB */
#if PARALLEL==1
      parallelizerdf(0);
#endif
      saverdf(rdf.grid); }

    if (icyc) {
      backup("sta");
      StaKeySave(Fn("sta"),stoptime);
      if (option('y')) {
        if (dt.plb==0) writeplayback();
        closeplayback();
        plbopened=0; }
      saveCP(icyc,stoptime); }

#ifdef POLAR
#include "polarres.c"
#endif /*# POLAR */

    /* release the heap */
    StaFree();
    release(mark);
#ifdef WIDOM
    widomrdf(-1);
#endif

    if (No.c) {
      int sp;

      loop (sp,0,nspec)
        if (constrit[sp].nit[COR_LAST])
          prt("spec=%d: omegac=%f (%s) last niter/molec=%f\n(cf. \"constrit%d spec%d\")",
            sp,constrit[sp].omegac,
            omegac<0?"optimized":"fixed",
            (double)constrit[sp].nit[COR_LAST]/spec[sp]->N,sp,0); }

    prt("sweep finished: L=(%.9f %.9f %.9f) V=%.11g",
                            VARG(box.L),      box.L[0]*box.L[1]*box.L[2]);

    fflush(out);

#endif /* !MARK */ /*# MARK */

    if (En.kin) put3(En.bonded,En.kin,En.bonded/En.kin)

    if (En.bonded>En.kin*10) {
      static int pass=0;
      if (!pass)
        WARNING(("Bonded energy is %g times greater than kinetic energy\n\
(more warnings will be suppressed)",En.bonded/En.kin))
      pass=1; }

    } /* >>> run >>> */

 TheEnd:

  if (sig>0) {
    if (sig==3) prt("quitted because nomax or stop reached");
    else if (sig==4) prt("quitted because of %s",Fn("stp"));
    else prt("quitted with sig=%d",sig); }

  if (fclose(out)) Error("close out");
  fclose(in);

  if (option('s') && option('b')) fputc('\a',stderr);

  return 0;
} /* main */
