#ifndef _SIMGLOB_
#  define _SIMGLOB_

/*
#  define SERIAL
for debugging only - serial loop instead of starting threads in parallel */

#  include "simopt.h"

#  ifdef GOLD
#    define WALL
#  endif /*# GOLD */

/* do not forget to #include "simopt.h" into simulation and project-specific
modules that do not #include "simglob.h" ! */

#  include "vector.h"

#  ifdef FLOAT
#    define realfmt "%f"
#  else /*# FLOAT */
#    define realfmt "%lf"
#  endif /*#!FLOAT */

#  ifndef PARALLEL
#    define PARALLEL 0
#  endif /*# PARALLEL */

#  if PARALLEL
#    include "sys4par.h"
#  endif /*# PARALLEL */

/* PRESSURETENSOR: sum of flags: */
#  define PT_VIR 1 /* virial (=configurational + constraint) part */
#  define PT_KIN 2 /* kinetic part */
#  define PT_ANY 3 /* virial or kinetic */
#  define PT_OFF 4 /* also off-diagonal terms (with PT_VIR and PT_KIN) */
#  define PT_MOL 8 /* also molecular-based (center-of-mass), modifies PT_KIN */
#  define PT_MOM 16 /* SPECIAL: 2nd and 3rd moments of center-of-mass energy */
#  ifndef PRESSURETENSOR
#    define PRESSURETENSOR 0
#  else /*# PRESSURETENSOR */
#    if PRESSURETENSOR == PT_OFF
#      error PRESSURETENSOR&PT_OFF (off-diagonal) and none of PRESSURETENSOR&PT_VIR (virial) and PRESSURETENSOR&PT_KIN (kinetic)
#    endif /*# PRESSURETENSOR == PT_OFF */
#    if PRESSURETENSOR == PT_MOL
#      error PRESSURETENSOR&PT_MOL (off-diagonal) and not PRESSURETENSOR&PT_KIN (kinetic)
#    endif /*# PRESSURETENSOR == PT_MOL */
#    if PRESSURETENSOR&PT_OFF
#      define PT_DIM (2*DIM)
#    else /*# PRESSURETENSOR&PT_OFF */
#      define PT_DIM (DIM)
#    endif /*#!PRESSURETENSOR&PT_OFF */
#    if (PRESSURETENSOR & PT_MOM) && !(PRESSURETENSOR&PT_MOL)
#      error PRESSURETENSOR: higher moments of CM velocity requested without CM velocity (PT_KIN)
#    endif
#  endif /*#!PRESSURETENSOR */

#  ifdef COOK
#    include "intrapot.h"
#  endif /*# COOK */

/*
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %                     basic control structures                      %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Species are numbered by integers from zero.
Information on each species is stored in a structure of type `specinfo_t' which
contains number of sites, masses and charges of sites, number of constraints
and corresponding sites and lengths.  These structures are of variable length
and there is an array of pointers 'spec' to these structures.
Example:  spec[1]->si[2].charge  is charge of site 2 of species 1

Any of N molecules (numbered from 0 to N-1) is described by a structure of
type `molecule_t' which contains the species number, number of sites and
constraints, information where to find the physical configuration (coordinates
of sites and, in Hamiltonian formalism, also momenta).  There is an array of
these molecule descriptors called `cfg'.

The physical configuration is stored in a special contiguous array of vectors
which is a part of a self-defined structure (`SDS', see "alloc.c") of type
`ToInt'.  There are several such self-defined structures containing higher
derivatives etc. as required by the Gear predictor-corrector method for solving
the differential equations.

The input information on interaction sites is stored in array of type
`sitedef_t' - see "simdef.c"

9/96: dependants:
   these are auxiliary sites whose positions are defined as linear
   combinations of other sites.
   Let
     depend_t *d;
     struct depitem_s *dep=d->dep;

   then
     r[d] =       SUM       r[dep[j].i] * dep[j].w
                j<d->n

   and analogously for the forces
9/2010: dependants extended to out-of-plane and code changed
*/

#  ifdef COOK
/* WARNING: this differs from blend: NBFIX solved there */
enum exc_e { END=-1,EXCL=0,ONEFOUR=1 };

typedef struct exception_s {
  int indx;   /* atom index */
  enum exc_e type;
  } exception_t;

#  endif /*# COOK */

typedef int pair_t[2];

typedef struct {
  real mass;      /* mass of site [extended! - more efficient, more memory] */
  real imass;     /* 1/mass of site */
  real charge;    /* charge of site */
  real bondq;     /* squared lengths of constraints (bonds) */
  pair_t pair;    /* site1 and site2 of constraint */
#  ifdef POLAR
  real chargepol; /* charge of auxiliary site to mimic polarisation */
  real qqpol;     /* chargepol^2 */
  real alpha_qq;  /* alpha/chargepol^2 where alpha=polarisability */
#    if POLAR&2
  real Esat;      /* saturation energy */
#    endif /*# POLAR&2 */
#    if POLAR&1
  real qkappa;    /* chargepol*kappa */
  int anion;      /* 1 1st of the shell-core pair (anion,kappa!=0) */
                  /* 2 2nd of the pair (cation, repulsive); param. rep */
#    endif /*# POLAR&1 */
#    if POLAR&4
  int  neutral;   /* !charge and !chargepol */
#    endif /*# POLAR&4 */
#    if POLAR&8
  real alphazz_qq;/* alphazz/chargepol^2 (z direction) */
  int tozz;       /* -1: no axial polarizability,
                     >=0: atom defining the direction */
#    endif /*# POLAR&8 */
#  endif /*# POLAR */
  int st;         /* site type - not in old versions! */
#  ifdef COOK
  exception_t *exc;  /* pointer to the list of exceptions */
#  endif /*# COOK */
#  ifndef FREEBC
  real sfweight;  /* atom weight for structure factor, see also sitedef_t */
#  endif /*# FREEBC */
  } siteinfo_t;

struct depitem_s {
  int i;   /* independent site in species */
  real w;  /* weight */ };

typedef struct depend_s {
  struct depend_s *next;   /* next in the list */
  int indx;                /* the dependent site */
  int n;                   /* n>0: "Middle" (linear) dependants:
                              number of reference sites (parents)
                              n=-3: "Lone" (general 3D - nonplanar) */
  struct depitem_s dep[3]; /* [|n|]
                              indices and weights of the parents
                              assumed: SUM dep[].w = 1 */
  /* Torque dependants (out of plane)
     local coord. system = (X,Y,Z), Z=X x Y (x=cross product) 
     f_x=X.f, etc */
  vector x;                /* X from 3 sites: X=sum_i x[i] r[i] */
  vector y;                /* Y from 3 sites: Y=sum_i y[i] r[i] */
  real wz;                 /* z-component of the site */ 
  vector tx;               /* tx[a]*f_x added to f[a]_z */
  vector ty;               /* ty[a]*f_y added to f[a]_z */
  } depend_t;

typedef struct {
  int ns;           /* # of sites */
  int nc;           /* # of constraints */
  int N;            /* # of molecules */
  int config;       /* set if abs init config (from blend) */
#  ifdef COOK
  int intra;        /* if the molecule has at least one bond,angle,torsion */
  int pot;          /* 0=default, 3=TIP3P, 4=TIP4P, 5=ST2 */
  bond_t *bond;     /* bonds [external constraints] table head */
  angle_t *angle;   /* angle table head */
  torsion_t *dihedral,*improper,*aromatics; /* heads of torsion tables */
  depend_t *dependants; /* table of lin. dependent sites */
  double zero_energy; /* const to be added to the pot energy */
  double charge;    /* molecular charge */
  double mass;      /* molecular mass */
  char *name;       /* species name */
#  endif /*# COOK */
  siteinfo_t si[1]; /* [max{ns,nc}] */
  } specinfo_t;

typedef struct {
  int sp;    /* species */
  int ns;    /* # of sites */
  int nc;    /* # of constraints */
  int ir;    /* relative address of 1st site in rp - in bytes (use rof) */
#  ifdef Hamilton
  int ip;    /* relative address of 1st momentum of site in rp (use pof) */
#  endif /*# Hamilton */
  int ig;    /* relative address of 1st g[amma] in gs */
#  ifdef ANCHOR
  int anchor;  /* OR of the following bits: */
                         /* MEASURE and PRINT (no constraint): */
#    define ANCHOR_r   0x01  /* position of center-of-mass */
#    define ANCHOR_v   0x02  /* velocity of center-of-mass */
#    define ANCHOR_x   0x04  /* acceleration of center-of-mass */
#    define ANCHOR_f   0x08  /* force to center-of-mass */
#    define ANCHOR_m   0x10  /* momentum of force (torque) to center-of-mass */

                         /* CONSTRAIN ("anchor" = keep constant): */
#    define ANCHOR_s  0x100  /* site (at r0) */
#    define ANCHOR_c  0x200  /* center-of-mass */
#    define ANCHOR_i  0x400  /* principial axis of the inertia tensor
                            (parallel to vector axis) */
#    define ANCHOR_a  0x800  /* axis given by center-of-mass and site */
#    define ANCHOR_p 0x1000  /* axis given by a pair of sites */
#    define ANCHOR_t 0x2000  /* axis perpendicular to a plane of 3 sites */
#    define ANCHOR_g 0x4000  /* center-of-mass of group(s) */
#    define ANCHOR_pos (ANCHOR_s|ANCHOR_c)
                         /* constrain any position BUT group CM */
#    define ANCHOR_axis (ANCHOR_i|ANCHOR_a|ANCHOR_p|ANCHOR_t)
                         /* constrain any axis */
  struct anchorgroup_s {
    struct anchorgroup_s *next;
    vector r0; /* the anchor point */
    vector cf; /* constrain force measured here */
    int ns;
    int site[1]; /* [ns] */ } *group;
  vector r0;   /* the anchor point for ANCHOR_{s,c} */
  int iaxis[3];/* site numbers for ANCHOR_{i,a,p,t}; iaxis[0] for ANCHOR_s */
  vector axis; /* the axis direction (normalized; ANCHOR_{i,a,p,t} only) */
#  endif /*# ANCHOR */
  } molecule_t;

/***
Type of function computing forces between m1 and m2.
It should return energy.
Forces are summed to f1 and f2.
Used by peo and cook
***/
typedef double pot2_t(
  vector f1[/*m1->ns*/], vector f2[/*m2->ns*/],
  molecule_t *m1, molecule_t *m2,
  vector *rp);

/***
Type of function computing intramolecular forces of molecule m and eventually
correcting for wrong site-site forces.  It should return energy.
Forces are summed to f.
Used if pairs are calculated by the link-cell method.
***/
typedef double pot1_t(
  vector f[/*m1->ns*/],
  molecule_t *m,
  vector *rp);

#  ifndef LINKCELL
#    define pot_t pot2_t
#  endif /*# LINKCELL */

typedef struct /*SDS*/ {
  int  size;    /* whole struct in bytes */
  int  padd;    /* padded to 8 bytes */
  real logs;    /* log of the Nose variable s */
  vector rp[1]; /* contiguous array of all r[ns] (and p[ns]) */
  } *ToIntPtr,ToInt;
/* WARNING: better DO NOT change this struct, dirty tricks used inside cook
   -- logs is the 1st real variable to be integrated
   -- reals must follow until the end of the structure
   -- MUST begin with 2 ints */

/***
    Macro  rof(M,RP)  returns the address of array of vectors containing
    positions of sites of molecule *M in array RP of all sites and momenta
    (typically RP=a[i]->rp).  In other words,  rof(M,RP)[i]  is a vector
    of 3 Cartesian coordinates of the i-th site.
    Macro  pof(M,RP)  does the same for momenta.
    POLAR: see also polarrof
***/
#  define rof(M,RP) ((vector*)((char*)(RP)+(M)->ir))
#  ifdef Hamilton
#    define pof(M,RP) ((vector*)((char*)(RP)+(M)->ip))
#  endif /*# Hamilton */

typedef struct {
  int b;          /* 2nd subscript, b<=a */
  real imass;     /* 1/mass of the common site - not def for a==b */
  real Mab;       /* values of M[a,b] = sum{i} 1/mi*dfa/di*dfb/di */
  } M_t;

typedef struct {
  double mass;    /* total mass */
  double free_mass; /* mass in all movable molecules */
  double charge;  /* total charge */
#  ifdef LINKCELL
  double percell; /* av. # of sites/cell */
  int cell[DIM];  /* # of cells per box */
#  endif /*# LINKCELL */
#  ifdef POLAR
  double A;       /* sum of all polarizabilities (in A^3), without 4PI */
  int pol;        /* # of dipole sites (used for Car-Parrinello) */
  int free_pol;   /* as above, in movable molecules
                     BUG: fixed molecules are not polarizable */
#  endif /*# POLAR */
#  if PARALLEL
  int th;         /* number of threads */
#    if PARALLEL==1 || PARALLEL==3
  int lcth;       /* number of threads for linked-cell */
#    endif /*# PARALLEL==1 || PARALLEL==3 */
  pthread_t *thread; /* [No.th]: pthread-style threads */
  pthread_attr_t thread_attr; /* thread attributes */
  int measureserial; /* force measurements in serial (void now) */
#  endif /*# PARALLEL */
  int N;          /* total # of molecules */
  int s;          /* total number of sites (sum over all molecules) */
  int c;          /* total # of constraints */
  int depend;     /* total # of dependants */
  int free_N;     /* # of movable molecules */
  int free_s;     /* # of sites in movable molecules */
  int free_c;     /* # of constraints in movable molecules */
  int free_depend;/* total # of dependants in movable molecules */
  int maxc;       /* max # of constraints / molecule */
  int maxs;       /* max # of sites / molecule */
  int f;          /* effective # of degrees of freedom */
  int f_tr;       /* effective # of intermolecular degrees of freedom */
  int f_in;       /* effective # of intramolecular degrees of freedom */
  int nreal;      /* # of reals in vectors (POLAR: 2*) */
  int eq;         /* # of equations of motion */
  int first;      /* # of molecules to be included in En.first */
  int pred;       /* option('p')%10 : predictor of constraints or velocity */
  int lone;       /* total # of Lone dependants */
  } No_t;

typedef struct {
  double T;      /* Nose correlation time or friction thermostat time */
  double E;      /* correlation time to keep energy constant */
  double rho;    /* typical time for density changes to reach given density */
/*  double T_tr,T_in;  for separate inter/intra molecular motions */
  double P;      /* isobaric friction-like correlation time */
/*  double s;    damping of ds/dt (equilibrating only!) (???) */
/*  double Q;    Nose constant - no longer used ! tau.Q = No.f*T*sqr(tau.T) */
#  ifdef POLAR
  double dip;    /* for Lagrange (Car-Parrinello) mechanical dipoles */
#  endif /*# POLAR */
  double R;      /* correlation time to change parameter LJ.RvdW in setss(0,i,i) */
  int i;         /* site (both sites must be the same) */
  } tau_t;

/*** En_t: summary of cycle statistics ***/

typedef struct {
  double pot;      /* potential energy, as sum of bond,LJ,el etc. */
  double kin;      /* kinetic energy (first calculated as 2*kin.en., later
                      halved to get the true kin.en.) */
  double tot;      /* total energy (pot+kin+zero_energy) */
  double el;       /* electrostatic energy; for Ewald both k- and r-space */
  double corr;     /* LJ cutoff correction: Ecorr=corr/V, Pcorr=corr/V^2 */
  double virc;     /* virial of constraint forces */
  double LJ;       /* Lennard-Jones energy (sum over all pairs) */
  double bonded;   /* sum of bond+angle+torsion terms */
  double intra;    /* site-site intramolecular terms (?) */
  double fix;      /* extra terms to fix some atom positions */
  double pot0,el0,LJ0,bonded0;  /* as above - No.first molecules */
  double potX,elX,LJX;          /* as above - all but No.first molecules */
  double LJmn;     /* mol-mol LJ energy, passed from charmm2 */
  double kin_tr;   /* 2*translational kin. energy (from center-of-mass) */
  double kin_in;   /* 2*intramolecular+rotatioal energy */
#  ifdef POLAR
  double self;     /* self energy */
  double kinpol,Tpol; /* mech.dip.: kin.energy and temperature */
  double polmaxerr,polstderr,polmaxdr;
                   /* errors during predictor/corrector or iterations */
  double Polmaxerr,Polstderr;
                   /* errors from measureP(); use short dV */
  int polnit;
#  endif /*# POLAR */
  double vir;      /* virial of force */
  double P;        /* virial pressure [Pa] */
  double PdV;      /* pressure from virtual V change [Pa] */
  double T;        /* kinetic temperature */
#  if PRESSURETENSOR&PT_VIR
  double Pvir[PT_DIM];  /* xx yy zz [yz zx xy] (virial part) */
#  endif /*# PRESSURETENSOR&PT_VIR */
#  if PRESSURETENSOR&PT_KIN
  double Pkin[PT_DIM];  /* xx yy zz [yz zx xy] (kinetic part - site-based) */
#  endif /*# PRESSURETENSOR&PT_KIN */
#  if PRESSURETENSOR&PT_MOL
  double PKin[PT_DIM];  /* xx yy zz [yz zx xy] (kinetic part - CM-based) */
#  endif /*# PRESSURETENSOR&PT_MOL */
#  if PRESSURETENSOR&PT_MOM
  double PKin2[DIM];  /* SPECIAL: PKin^2 (to calc. 2nd moment = kurtosis of v) */
  double PKin3[DIM];  /* SPECIAL: PKin^3 (to calc. 3rd moment) */
#  endif /*# PRESSURETENSOR&PT_MOM */
#  if PRESSURETENSOR
  double Ptens[PT_DIM];  /* xx yy zz [yz zx xy] total */
#  endif /*# PRESSURETENSOR */
  double T_in;     /* kin. T from intramolecular motions and rotations*/
  double T_tr;     /* kin. T from intermolecular motions (translations) */
  double r1;       /* error of constraints of predictor */
  double r2;       /* error of constraints after integration step */
  double r3;       /* error of constraints (of true configuration) */
#  ifdef Lagrange
  double v1,v2,v3; /* errors of velocity constraints */
#  endif /*# Lagrange */
#  ifdef SHEAR
  double Cv;       /* SUM v_i[xy] cos(2PIz/L) / SUM cos^2(2PIz/L) */
  double kinshear; /* kin E of shear flow */
#  endif /*# SHEAR */
  double cmdrift;   /* CM position */
  double vdrift;    /* velocity of CM */
  double angdrift;  /* angular velocity */
  vector J;         /* current density in SI [A/m2] */
  double logs,dlogs; /* a[0]->logs, a[1]->logs */
  } En_t;

#  include "rdf.h"

/****** miscellaneous ******/
/* for some strange compatibility reasons,
   we do not #include "options.h" here */
#  ifndef option
extern int optionlist[32];
#    define option(X) optionlist[X & 31]
#  endif /*# option */

#  ifdef COOK
#    define FROM option('j')
#  else /*# COOK */
#    define FROM 0
#  endif /*#!COOK */

/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                     declarations of global variables                  %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

extern
pot2_t ***pot;       /* array [nspec][nspec] of pointers to functions computing
                        INTERmolecular interactions */

#  ifdef LINKCELL
extern
real over14;         /* factor by which the estimate of change of the max
                        1-2,1-3,1-4 distance is multiplied.
                        Must be >1; the higher, the safer
                        negative: also print distances and speeds
                        1.5 is pessimistic for proteins,
                        but just enough for water */
extern
pot1_t  **pot1;      /* array [nspec] of pointers to functions computing
                        INTRAmolecular interactions */
#  endif /*# LINKCELL */

extern
specinfo_t **spec;   /* array [nspec] of pointers to species definitions */

extern
M_t **Ms;            /* array [nspec or N] of pointers to matrices M */

extern
molecule_t *cfg;     /* array [N] of molecule descriptors */
extern
ToIntPtr a[9];       /* [MaxGearOrder+3] */
extern
ToIntPtr fixa;       /* fixed sites (by harmonic springs) */

/* warning: even #ifdef FLOAT, the constraint calculations are in double */
extern
double **gs;         /* [option('p')] tables of Lagrange multipliers g to
                        predict their values
                     NOTE: pointers are NOT constant because swaping of cor-
                     responding arrays is implemented by pointer interchange */
extern
double *binom;       /* [option('p')] coefficients of this predictor */

extern
rdf_t ***rdf;        /* table of pointers to rdf's
                        see rdf.h for rdf_t
                        note: also in each sitesite_t sstab[][] (see setss) */

extern int init;        /* INITIAL KEY :
__init____.cfg___.cp____.sta .rdf .dih__logs
    0     read  append      read       read
    1     read  append      new        read
    2     read  rewrite     new         0
   >=3    init  rewrite     new         0  (sequential version only)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
extern int init_append; /* = init but :=0 after one step (to control appending) */

extern No_t No;        /* # of sites, constraints, degrees of freedom etc. */

extern volatile double t;     /* MD time */
extern real h;       /* time step */

extern double eps;   /* accuracy for computing Lagrange multipliers
                      if >=1 then # of iteration steps */
extern double omega; /* not implemented */

extern double epsc;  /* accuracy for correcting constraints
                      if epsc>=1, then # of iteration steps
                      (default is changed for -c4) */
extern double omegac;/* relaxation parameter for 'correct' */

#  ifdef POLAR
extern double epsp;  /* accuracy for self-field, in pdu. (program dipole units)
                        (1 pdu.= 0.00245 eA = 0.01175 D) */
extern double epspx; /* as above, for P measure by virtual V change and similar
                        testing purposes */
extern double omegap;/* relaxation paramater for self-field iterations
                        see also option -^ */
extern int maxitp;   /* max # of self-field iterations until WARNING
                        (another maxitp until ERROR) */
#  endif /*# POLAR */

extern double T;     /* Nose|friction temperature (if tau.T) [K] */
extern double P;     /* pressure (friction, if tau.P) [Pa] */

extern En_t En;      /* energies, virial, pressure tensor, constraint errors.. */

extern int GearOrder;/* see also option('m') */
extern struct lag_s {
  int err;           /* lag for error analysis of important variables */
  int n;             /* number of blocked calculations: the max.block size is 2^n */
#  ifdef RGYR
  int dim;           /* how many coordinates recorded (3 for vx,vy,vz) */
  int nv;            /* how many molecules */
  int v;             /* lag for velocity-velocity time autocorr. functions */
#  endif /*# RGYR */
  int cond;          /* lag for current time autocorr.f. (conductivity) */
#  if PRESSURETENSOR&PT_OFF
  int visc;          /* lag for Ptxy,... time autocorr.f. (viscosity) */
#  endif /*# PRESSURETENSOR&PT_OFF */
} lag;

extern double E;     /* total energy in [J/mol] (=Eunit) to be kept constant */
extern tau_t tau;    /* relaxation times for thermostat, volume change,... */
extern double T_tr_in; /* ratio T_tr/T_in for thermostat=13 */
extern int thermostat;
#  define T_BERENDSEN   1
#  define T_FRICTION    1
#  define T_NOSE        2
#  define T_ANDERSEN    3 /* randomly selected atoms updated */
#  define T_MAXWELL     4 /* periodically all atoms updated at once */
#  define T_ANDERSEN_CM 5 /* randomly selected molecules (center-of-mass) updated */
#  define T_MAXWELL_CM  6 /* periodically all molecules updated at once */
#  define T_TR         11 /* friction only translations only */
#  define T_IN         12 /* friction only intramolecular velocities */
#  define T_FRICTIONS  13 /* combined 11,12: see T_tr_in */

extern int measure;  /* 0: only forces required
                        1: also energy etc.
                        2: special for calculating # of pairs (DEBUG only) */

/*** Ewald related global variables - see  comments in ewald.c and erfc.c ***/
/* they are used also for not-Ewald sums, so they are here... */

extern struct el_s {
  real alpha;      /* Ewald separation parameter in units of 1/A;
                      also cutoff electrostatic parameter */
  real kappa;      /* k-space cutoff in form kappa=K[i]/L[i] */
  real epsr,epsk;  /* expected r- and k-space errors in forces, in K/A */
  real minqq;      /* minimum expected charge-charge separation (affects
                      only reported error of approximation of erfc) */
  real epsinf;     /* dielectric constant at infinity (must be very big if
                      free charges are present */
  real diff;       /* max rel.change of L before a warning about "unchanged K" */
  int grid;        /* # of grid points per unity of squared distance */
  int test;        /* test module switch (former etest) */
  int diag;        /* whether to include diagonal correction */
} el;

extern struct box_s {
  real cutoff;  /* r-space cutoff */
  vector L;     /* size of the periodic box;
                   it's important that L is real, not double, for automatic
                   float<->double conversion in loadcfg() */
  real cq;      /* =cutoff^2 */
  vector Lh;    /* Lh=L/2 */
#  ifdef LINKCELL
  real rmin;    /* minimum x,y,z of (any site of) a molecule - user value */
  real r0;      /* minimum x,y,z of (any site of) a molecule - used */
#  endif /*# LINKCELL */
} box;

extern int cache;    /* molecules in block for r-space sum */

/* Q_t *Q; is in ewald.h and ewald.c */

#  ifdef POLAR
extern
int polar_off; /* offset of auxiliary charges in the units of bytes */
#    define polarrof(M,RP) ((vector*)((char*)(RP)+(M)->ir+polar_off))
#  endif /*# POLAR */

#  ifdef SHEAR
extern double shear; /* max gradient of acceleration for cos(z) method */
#  endif /*# SHEAR */

typedef struct fixsites_s {
  struct fixsites_s *next;
  int from,to; /* C-style range of sites: [from,to) */
  int isr; /* whether r applies */
  vector r; /* position */
  } fixsites_t;
extern fixsites_t *fixsites0;

extern int drift;
#  define DRIFT_X 1 /* center-of-mass */
#  define DRIFT_Y 2
#  define DRIFT_Z 4
#  define DRIFT_VX 8 /* momentum (velocity) */
#  define DRIFT_VY 16
#  define DRIFT_VZ 32
#  define DRIFT_AX 64 /* angular momentum drift */
#  define DRIFT_AY 128
#  define DRIFT_AZ 256
#  define DRIFT_WHEN 1536
#  define DRIFT_CYCLE 0 /* every cycle, init, load, save */
#  define DRIFT_START 512 /* init, load only, then cancel */
#  define DRIFT_STEP  1024 /* every step, init, load, save */
#  define DRIFT_SAVE  1536 /* init, load, save */
#  define DRIFT_DEPEND 2048 /* dependants every step (former norm&4) */
#  define DRIFT_AUTO 4096 /* auto set */
/* WARNING: do not change numbers - coded sometimes directly by numbers */
extern int conserved; /* number of conserved degrees of freedom */

#  define NCENTER 10 /* number of force terms */

extern struct center_s {
#  ifdef FREEBC
  double r0,K;
#  else /*# FREEBC */
  int on;              /* flag: 2=to box center, 1=z-pos */
  vector K,K2;         /* force constant to center-of-box */
  vector r0;           /* range (force starts at center+-r0) */
#    ifdef SLAB
  int n[NCENTER];      /* how many molecules is affected (input) */
  int nn[NCENTER];     /* molecules affected [range) */
  int ns[NCENTER];     /* affected sites: 0: all, >0: i<ns only, <0: i>=|ns| */
  double z[NCENTER];   /* center relatively to Lh[2] */
  double z0[NCENTER];  /* radius of zero-pot. range */
  double Kz[NCENTER];  /* z-force const, for molecules up to n[] */
  double Kz2[NCENTER]; /* z1=0: 2*Kz, z1!=0: step */
  double z1[NCENTER];  /* for slab/outside bias function (input) */
  double dz[NCENTER];  /* z1-z0, for z1!=0 */
#    endif /*# SLAB */
#  endif /*#!FREEBC */
  vector cmK;          /* global force of the center-of-mass to (box) center */
  double cmmass;       /* mass of cmn molecules (from FROM) */
  int cmn;             /* # of molecules affected; 0=all but FROM */
  } center;

struct Eelst_s {
  int isE;             /* 1 if there is nonzero E */
  vector E;            /* in prog. units */
  } Eext;

#  ifdef ANCHOR
extern struct anchor_s {
  FILE *f; /* output file of anchor forces .anc */
  double rr; /* max displacement, squared */
  double cos,sin; /* max cos,sin of rot angle */
  } anchor;
#  endif /*# ANCHOR */

extern int onefourinrdf; /* control whether 1-4 included in RDF's */

#  if PARALLEL
  /* usage of the pthread-based parallel macro:

     PARALLELIZE(FUNCTION)

     will call option('\\')-times function FUNCTION, which is passed
     the thread number in the form of void* casted int*
     (thread 0 is the caller)

     Example:

     void *FUNCTION(void *arg)
     {
       int ithread=*(int*)arg;
       fprintf(stderr,"thread %d started\n",ithread);
       ...
       return arg;
     }
  */

#    define PARALLELIZE(FUNC,NTHREADS) { \
  int ith; char *rc; \
  loop (ith,1,NTHREADS) \
    if (pthread_create(No.thread+ith,&No.thread_attr,FUNC,(void*)(No.thread+ith))) \
      ERROR(("cannot create thread %d",ith)) \
  ith=0; \
  FUNC((void*)No.thread); \
  loop (ith,1,NTHREADS) \
    if (pthread_join(No.thread[ith],(void**)&rc)) \
    ERROR(("thread %d failed",ith)) }

#    ifdef SERIAL
/* serial version for debugging: */
#      define PARALLELIZE(FUNC,NTHREADS) { \
  int ith; char *rc; \
  loop (ith,0,NTHREADS) FUNC((void*)(No.thread+ith)); }
#    endif /*# SERIAL */

#  endif /*# PARALLEL */

#endif /*# _SIMGLOB_ */
